import abc
import json
from pathlib import Path
from typing import Any, Optional

import requests
from fcache.cache import FileCache as _FileCache
from UnleashClient.constants import FEATURES_URL, REQUEST_TIMEOUT


class BaseCache(abc.ABC):
    """
    Abstract base class for caches used for UnleashClient.

    If implementing your own bootstrapping methods, you must set the `bootstrapped` attribute to True after configuration is set.
    """
    bootstrapped = False

    @abc.abstractmethod
    def set(self, key: str, value: Any):
        pass

    @abc.abstractmethod
    def mset(self, data: dict):
        pass

    @abc.abstractmethod
    def get(self, key: str, default: Optional[Any] = None):
        pass

    @abc.abstractmethod
    def exists(self, key: str):
        pass

    @abc.abstractmethod
    def destroy(self):
        pass


class FileCache(BaseCache):
    """
    The default cache for UnleashClient.  Uses `fcache <https://pypi.org/project/fcache/>`_ behind the scenes.

    You can boostrap the FileCache with initial configuration to improve resiliency on startup.  To do so:

    - Create a new FileCache instance.
    - Bootstrap the FileCache.
    - Pass your FileCache instance to UnleashClient at initialization along with `boostrap=true`.

    You can bootstrap from a dictionary, a json file, or from a URL.  In all cases, configuration should match the Unleash `/api/client/features <https://docs.getunleash.io/api/client/features>`_ endpoint.

    Example:

    .. code-block:: python

        from pathlib import Path
        from UnleashClient.cache import FileCache
        from UnleashClient import UnleashClient

        my_cache = FileCache("HAMSTER_API")
        my_cache.bootstrap_from_file(Path("/path/to/boostrap.json"))
        unleash_client = UnleashClient(
            "https://my.unleash.server.com",
            "HAMSTER_API",
            cache=cache
        )

    :param name: Name of cache.
    :param directory: Location to create cache.  If empty, will use filecache default.
    """
    def __init__(self, name: str, directory: Optional[str] = None):
        self._cache = _FileCache(name, app_cache_dir=directory)

    def bootstrap_from_dict(self, initial_config: dict) -> None:
        """
        Loads initial Unleash configuration from a dictionary.

        Note: Pre-seeded configuration will only be used if UnleashClient is initialized with `bootstrap=true`.

        :param initial_config: Dictionary that contains initial configuration.
        """
        self.set(FEATURES_URL, initial_config)
        self.bootstrapped = True

    def bootstrap_from_file(self, initial_config_file: Path) -> None:
        """
        Loads initial Unleash configuration from a file.

        Note: Pre-seeded configuration will only be used if UnleashClient is initialized with `bootstrap=true`.

        :param initial_configuration_file: Path to document containing initial configuration.  Must be JSON.
        """
        with open(initial_config_file, "r",  encoding="utf8") as bootstrap_file:
            self.set(FEATURES_URL, json.loads(bootstrap_file.read()))
            self.bootstrapped = True

    def bootstrap_from_url(self, initial_config_url: str, headers: Optional[dict] = None) -> None:
        """
        Loads initial Unleash configuration from a url.

        Note: Pre-seeded configuration will only be used if UnleashClient is initialized with `bootstrap=true`.

        :param initial_configuration_url: Url that returns document containing initial configuration.  Must return JSON.
        :param headers: Headers to use when GETing the initial configuration URL.
        """
        response = requests.get(initial_config_url, headers=headers, timeout=REQUEST_TIMEOUT)
        self.set(FEATURES_URL, response.json())
        self.bootstrapped = True

    def set(self, key: str, value: Any):
        self._cache[key] = value
        self._cache.sync()

    def mset(self, data: dict):
        self._cache.update(data)
        self._cache.sync()

    def get(self, key: str, default: Optional[Any] = None):
        return self._cache.get(key, default)

    def exists(self, key: str):
        return key in self._cache

    def destroy(self):
        return self._cache.delete()



import logging
import sys
import time

from UnleashClient import UnleashClient
from UnleashClient.strategies import Strategy


# ---
class DogTest(Strategy):
    def load_provisioning(self) -> list:
        return [x.strip() for x in self.parameters["sound"].split(',')]

    def apply(self, context: dict = None) -> bool:
        """
        Turn on if I'm a dog.

        :return:
        """
        default_value = False

        if "sound" in context.keys():
            default_value = context["sound"] in self.parsed_provisioning

        return default_value


# ---

root = logging.getLogger()
root.setLevel(logging.DEBUG)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
root.addHandler(handler)
# ---

custom_strategies_dict = {
    "amIADog": DogTest,
}

my_client = UnleashClient(
    url="https://app.unleash-hosted.com/demo/api",
    environment="staging",
    app_name="pyIvan",
    custom_headers={'Authorization': '56907a2fa53c1d16101d509a10b78e36190b0f918d9f122d'},
    custom_strategies=custom_strategies_dict,
    verbose_log_level=10
)

my_client.initialize_client()

while True:
    time.sleep(10)
    context = {
        'userId': "1",
        'sound': 'woof'
    }
    print(f"ivantest: {my_client.is_enabled('ivantest', context)}")
    print(f"ivan-variations: {my_client.get_variant('ivan-variations', context)}")
    print(f"ivan-customstrategyx: {my_client.is_enabled('ivan-customstrategy', context)}")



import uuid
import json
import pytest
import responses
from UnleashClient import UnleashClient
from UnleashClient.constants import FEATURES_URL, METRICS_URL, REGISTER_URL
from tests.utilities.testing_constants import URL, APP_NAME


MOCK_JSON = """
{
  "version": 2,
  "features": [
      {
          "name": "F8.semverEQ",
          "description": "semver",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "version",
                          "operator": "SEMVER_EQ",
                          "value": "1.2.2"
                      }
                  ]
              }
          ]
      },
      {
          "name": "F8.semverGT",
          "description": "semver",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "version",
                          "operator": "SEMVER_GT",
                          "value": "1.2.2"
                      }
                  ]
              }
          ]
      },
      {
          "name": "F8.semverLT",
          "description": "semver",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "version",
                          "operator": "SEMVER_LT",
                          "value": "1.2.2"
                      }
                  ]
              }
          ]
      },
      {
          "name": "F8.semverAlphaGT",
          "description": "semver",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "version",
                          "operator": "SEMVER_GT",
                          "value": "2.0.0-alpha.1"
                      }
                  ]
              }
          ]
      },
      {
          "name": "F8.semverAlphaLT",
          "description": "semver",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "version",
                          "operator": "SEMVER_LT",
                          "value": "2.0.0-alpha.1"
                      }
                  ]
              }
          ]
      },
      {
          "name": "F8.semverAlphaVersioning",
          "description": "semver",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "version",
                          "operator": "SEMVER_LT",
                          "value": "2.0.0-alpha.3"
                      }
                  ]
              }
          ]
      },
      {
          "name": "F8.alphaUnnumbered",
          "description": "semver",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "version",
                          "operator": "SEMVER_GT",
                          "value": "2.0.0-alpha"
                      }
                  ]
              }
          ]
      },
      {
          "name": "F8.releaseCandidate",
          "description": "semver",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "version",
                          "operator": "SEMVER_GT",
                          "value": "2.0.0-rc"
                      }
                  ]
              }
          ]
      }
  ]
}
"""


@pytest.fixture()
def unleash_client():
    unleash_client = UnleashClient(url=URL,
                                   app_name=APP_NAME,
                                   instance_id='pytest_%s' % uuid.uuid4())
    yield unleash_client
    unleash_client.destroy()


@responses.activate
def test_feature_constraintoperators_semvereq_enabled(unleash_client):
    """
    F8.semverEQ should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F8.semverEQ", {'version': '1.2.2'})


@responses.activate
def test_feature_constraintoperators_semvereq_disabled(unleash_client):
    """
    F8.semverEQ should be disabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F8.semverEQ", {'version': '1.2.0'})


@responses.activate
def test_feature_constraintoperators_semvergt_enabled(unleash_client):
    """
    F8.semverGT should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F8.semverGT", {'version': '1.2.3'})


@responses.activate
def test_feature_constraintoperators_semvergt_disabled(unleash_client):
    """
    F8.semverGT should be disabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F8.semverGT", {'version': '1.2.0'})


@responses.activate
def test_feature_constraintoperators_semverlt_enabled(unleash_client):
    """
    F8.semverLT should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F8.semverLT", {'version': '1.2.1'})


@responses.activate
def test_feature_constraintoperators_semverlt_disabled(unleash_client):
    """
    F8.semverLT should be disabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F8.semverLT", {'version': '1.2.3'})


@responses.activate
def test_feature_constraintoperators_semveralphagt_beta(unleash_client):
    """
    F8.semverAlphaGT is less than beta
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F8.semverAlphaGT", {'version': '2.0.0-beta.1'})


@responses.activate
def test_feature_constraintoperators_semveralphagt_release(unleash_client):
    """
    F8.semverAlphaGT is less than release
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F8.semverAlphaGT", {'version': '2.0.0'})


@responses.activate
def test_feature_constraintoperators_semveralphalt_beta(unleash_client):
    """
    F8.semverAlphaLT is less than beta
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F8.semverAlphaLT", {'version': '2.0.0-beta.1'})


@responses.activate
def test_feature_constraintoperators_semveralphalt_oldbeta(unleash_client):
    """
    F8.semverAlphaLT is greater than old beta
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F8.semverAlphaLT", {'version': '1.9.1-beta.1'})


@responses.activate
def test_feature_constraintoperators_semveralphalt_release(unleash_client):
    """
    F8.semverAlpha is less than release
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F8.semverAlphaLT", {'version': '2.0.0'})


@responses.activate
def test_feature_constraintoperators_semveralphalt_alphas_lt(unleash_client):
    """
    F8.semverAlphaVersioning alpha.1 is less than alpha.3
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F8.semverAlphaVersioning", {'version': '2.0.0-alpha.1'})


@responses.activate
def test_feature_constraintoperators_semveralphalt_dif_alphas_gt(unleash_client):
    """
    F8.semverAlphaVersioning alpha.4 is greater than alpha.3
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F8.semverAlphaVersioning", {'version': '2.0.0-alpha.4'})


@responses.activate
def test_feature_constraintoperators_semverunnumbered_lt(unleash_client):
    """
    "F8.alphaUnnumbered - unnumbered is LT than numbered
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F8.alphaUnnumbered", {'version': '2.0.0-alpha.1'})


def test_feature_constraintoperators_semverrc_alpha(unleash_client):
    """
    F8.releaseCandidate - alpha is not greater than rc
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F8.releaseCandidate", {'version': '2.0.0-alpha.1'})

def test_feature_constraintoperators_semverc_beta(unleash_client):
    """
    F8.releaseCandidate - beta is not greater tha rc
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F8.releaseCandidate", {'version': '2.0.0-beta.1'})


def test_feature_constraintoperators_semverc_release(unleash_client):
    """
    F8.releaseCandidate - release is greater than rc
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F8.releaseCandidate", {'version': '2.0.0'})



VARIANTS = \
    [
        {
            "name": "VarA",
            "weight": 34,
            "payload": {
                "type": "string",
                "value": "Test1"
            },
            "overrides": [
                {
                    "contextName": "userId",
                    "values": [
                        "1"
                    ]
                }
            ]
        },
        {
            "name": "VarB",
            "weight": 33,
            "payload": {
                "type": "string",
                "value": "Test 2"
            }
        },
        {
            "name": "VarC",
            "weight": 33,
            "payload": {
                "type": "string",
                "value": "Test 3"
            }
        }
    ]

VARIANTS_WITH_STICKINESS = \
    [
        {
            "name": "VarA",
            "weight": 34,
            "stickiness": "customField",
            "payload": {
                "type": "string",
                "value": "Test1"
            },
            "overrides": [
                {
                    "contextName": "userId",
                    "values": [
                        "1"
                    ]
                }
            ]
        },
        {
            "name": "VarB",
            "weight": 33,
            "stickiness": "customField",
            "payload": {
                "type": "string",
                "value": "Test 2"
            }
        },
        {
            "name": "VarC",
            "weight": 33,
            "stickiness": "customField",
            "payload": {
                "type": "string",
                "value": "Test 3"
            }
        }
    ]



from typing import Tuple
import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from UnleashClient.constants import REQUEST_TIMEOUT, FEATURES_URL, \
    REQUEST_RETRIES
from UnleashClient.utils import LOGGER, log_resp_info


# pylint: disable=broad-except
def get_feature_toggles(url: str,
                        app_name: str,
                        instance_id: str,
                        custom_headers: dict,
                        custom_options: dict,
                        project: str = None,
                        cached_etag: str = '') -> Tuple[dict, str]:
    """
    Retrieves feature flags from unleash central server.

    Notes:
    * If unsuccessful (i.e. not HTTP status code 200), exception will be caught and logged.
      This is to allow "safe" error handling if unleash server goes down.

    :param url:
    :param app_name:
    :param instance_id:
    :param custom_headers:
    :param custom_options:
    :param project:
    :param cached_etag:
    :return: (Feature flags, etag) if successful, ({},'') if not
    """
    try:
        LOGGER.info("Getting feature flag.")

        headers = {
            "UNLEASH-APPNAME": app_name,
            "UNLEASH-INSTANCEID": instance_id
        }

        if cached_etag:
            headers['If-None-Match'] = cached_etag

        base_url = f"{url}{FEATURES_URL}"
        base_params = {}

        if project:
            base_params = {'project': project}

        adapter = HTTPAdapter(max_retries=Retry(total=REQUEST_RETRIES, status_forcelist=[500, 502, 504]))
        with requests.Session() as session:
            session.mount("https://", adapter)
            session.mount("http://", adapter)
            resp = session.get(base_url,
                               headers={**custom_headers, **headers},
                               params=base_params,
                               timeout=REQUEST_TIMEOUT, **custom_options)

        if resp.status_code not in [200, 304]:
            log_resp_info(resp)
            LOGGER.warning("Unleash Client feature fetch failed due to unexpected HTTP status code.")
            raise Exception("Unleash Client feature fetch failed!")

        etag = ''
        if 'etag' in resp.headers.keys():
            etag = resp.headers['etag']

        if resp.status_code == 304:
            return None, etag

        return resp.json(), etag
    except Exception as exc:
        LOGGER.exception("Unleash Client feature fetch failed due to exception: %s", exc)

    return {}, ''



from typing import Optional
from UnleashClient.features.Feature import Feature
from UnleashClient.variants.Variants import Variants
from UnleashClient.constants import FEATURES_URL, FAILED_STRATEGIES
from UnleashClient.utils import LOGGER
from UnleashClient.cache import BaseCache


# pylint: disable=broad-except
def _create_strategies(provisioning: dict,
                       strategy_mapping: dict,
                       cache: BaseCache,
                       global_segments: Optional[dict]) -> list:
    feature_strategies = []

    for strategy in provisioning["strategies"]:
        try:
            if "parameters" in strategy.keys():
                strategy_provisioning = strategy['parameters']
            else:
                strategy_provisioning = {}

            if "constraints" in strategy.keys():
                constraint_provisioning = strategy['constraints']
            else:
                constraint_provisioning = {}

            if "segments" in strategy.keys():
                segment_provisioning = strategy['segments']
            else:
                segment_provisioning = []

            feature_strategies.append(strategy_mapping[strategy['name']](
                constraints=constraint_provisioning, parameters=strategy_provisioning, global_segments=global_segments, segment_ids=segment_provisioning
            ))
        except Exception as excep:
            strategies = cache.get(FAILED_STRATEGIES, [])

            if strategy['name'] not in strategies:
                LOGGER.warning("Failed to load strategy. This may be a problem with a custom strategy. Exception: %s", excep)
                strategies.append(strategy['name'])

            cache.set(FAILED_STRATEGIES, strategies)

    return feature_strategies


def _create_feature(provisioning: dict,
                    strategy_mapping: dict,
                    cache: BaseCache,
                    global_segments: Optional[dict]) -> Feature:
    if "strategies" in provisioning.keys():
        parsed_strategies = _create_strategies(provisioning, strategy_mapping, cache, global_segments)
    else:
        parsed_strategies = []

    if "variants" in provisioning:
        variant = Variants(provisioning['variants'], provisioning['name'])
    else:
        variant = None

    return Feature(name=provisioning["name"],
                   enabled=provisioning["enabled"],
                   strategies=parsed_strategies,
                   variants=variant
                   )


def load_features(cache: BaseCache,
                  feature_toggles: dict,
                  strategy_mapping: dict,
                  global_segments: Optional[dict] = None) -> None:
    """
    Caching

    :param cache: Should be the cache class variable from UnleashClient
    :param feature_toggles: Should be the features class variable from UnleashClient
    :param strategy_mapping:
    :return:
    """
    # Pull raw provisioning from cache.
    feature_provisioning = cache.get(FEATURES_URL)
    if not feature_provisioning:
        LOGGER.warning("Unleash client does not have cached features. "
                       "Please make sure client can communicate with Unleash server!")
        return

    # Parse provisioning
    parsed_features = {}
    feature_names = [d["name"] for d in feature_provisioning["features"]]

    if "segments" in feature_provisioning.keys():
        segments = feature_provisioning["segments"]
        global_segments = {segment["id"]: segment for segment in segments}
    else:
        global_segments = {}

    for provisioning in feature_provisioning["features"]:
        parsed_features[provisioning["name"]] = provisioning

    # Delete old features/cache
    for feature in list(feature_toggles.keys()):
        if feature not in feature_names:
            del feature_toggles[feature]

    # Update existing objects
    for feature in feature_toggles.keys():
        feature_for_update = feature_toggles[feature]
        strategies = parsed_features[feature]["strategies"]

        feature_for_update.enabled = parsed_features[feature]["enabled"]
        if strategies:
            parsed_strategies = _create_strategies(parsed_features[feature], strategy_mapping, cache, global_segments)
            feature_for_update.strategies = parsed_strategies

        if 'variants' in parsed_features[feature]:
            feature_for_update.variants = Variants(
                parsed_features[feature]['variants'],
                parsed_features[feature]['name']
            )

    # Handle creation or deletions
    new_features = list(set(feature_names) - set(feature_toggles.keys()))

    for feature in new_features:
        feature_toggles[feature] = _create_feature(parsed_features[feature], strategy_mapping, cache, global_segments)



import pytest
import platform
from UnleashClient.strategies import ApplicationHostname


@pytest.fixture()
def strategy():
    yield ApplicationHostname(parameters={"hostNames": "%s,garbage,garbage2" % platform.node()})


def test_applicationhostname(strategy):
    assert strategy.execute()


def test_applicationhostname_nomatch():
    nomatch_strategy = ApplicationHostname(parameters={"hostNames": "garbage,garbage2"})
    assert not nomatch_strategy.execute()



import uuid
import json
import pytest
import responses
from UnleashClient import UnleashClient
from UnleashClient.constants import FEATURES_URL, METRICS_URL, REGISTER_URL
from tests.utilities.testing_constants import URL, APP_NAME


MOCK_JSON = """
{
  "version": 1,
  "features": [{
      "name": "Feature.A4",
      "description": "Enabled toggle for 100%",
      "enabled": true,
      "strategies": [{
        "name": "gradualRolloutSessionId",
        "parameters": {
          "percentage": "100",
          "groupId": "AB12A"
        }
      }]
    },
    {
      "name": "Feature.B4",
      "description": "Enabled toggle for 50%",
      "enabled": true,
      "strategies": [{
        "name": "gradualRolloutSessionId",
        "parameters": {
          "percentage": "50",
          "groupId": "AB12A"
        }
      }]
    },
    {
      "name": "Feature.C4",
      "enabled": true,
      "strategies": [{
        "name": "gradualRolloutSessionId",
        "parameters": {
          "percentage": "0",
          "groupId": "AB12A"
        }
      }]
    },
    {
      "name": "Feature.D4",
      "enabled": true,
      "strategies": [{
          "name": "gradualRolloutSessionId",
          "parameters": {
            "percentage": "0",
            "groupId": "AB12A"
          }
        },
        {
          "name": "default"
        }
      ]
    }
  ]
}
"""


@pytest.fixture()
def unleash_client():
    unleash_client = UnleashClient(url=URL,
                                   app_name=APP_NAME,
                                   instance_id='pytest_%s' % uuid.uuid4())
    yield unleash_client
    unleash_client.destroy()


@responses.activate
def test_feature_a4_enabled(unleash_client):
    """
    Feature.A4 should be enabled for user on context
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {
        "sessionId": "123"
    }

    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.A4", context_values)


@responses.activate
def test_feature_a4_nocontext(unleash_client):
    """
    Feature.A4 should be disabled when no user on context
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {}

    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.A4", context_values)


@responses.activate
def test_feature_b4_enabled(unleash_client):
    """
    Feature.B4 should be enabled for sessionId=122
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {
        "sessionId": "122"
    }

    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.B4", context_values)


@responses.activate
def test_feature_b4_disabled(unleash_client):
    """
    Feature.B4 should be disabled for sessionId=155
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {
        "sessionId": "155"
    }

    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.B4", context_values)


@responses.activate
def test_feature_c4(unleash_client):
    """
    Feature.C4 should be disabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {
        "sessionId": "122"
    }

    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.C4", context_values)


@responses.activate
def test_feature_d4(unleash_client):
    """
    Feature.D4 should be enabled for all because of default strategy
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {
        "sessionId": "122"
    }

    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.D4", context_values)



import uuid
from datetime import datetime
import json
import pytest
import responses
from UnleashClient import UnleashClient
from UnleashClient.constants import FEATURES_URL, METRICS_URL, REGISTER_URL
from tests.utilities.testing_constants import URL, APP_NAME


MOCK_JSON = """
{
  "version": 1,
  "features": [
      {
          "name": "F1.startsWith",
          "description": "startsWith",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "customField",
                          "operator": "STR_STARTS_WITH",
                          "values": ["some-string"]

                      }
                  ]
              }
          ]
      },
      {
          "name": "F2.startsWith.multiple",
          "description": "endsWith",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "customField",
                          "operator": "STR_STARTS_WITH",
                          "values": ["e1", "e2"]

                      }
                  ]
              }
          ]
      },
      {
          "name": "F3.endsWith",
          "description": "endsWith",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "email",
                          "operator": "STR_ENDS_WITH",
                          "values": ["@some-email.com"]
                      }
                  ]
              }
          ]
      },
      {
          "name": "F3.endsWith.ignoringCase",
          "description": "endsWith",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "email",
                          "operator": "STR_ENDS_WITH",
                          "values": ["@some-email.com"],
                          "caseInsensitive": true
                      }
                  ]
              }
          ]
      },
      {
          "name": "F4.contains",
          "description": "contains",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "email",
                          "operator": "STR_CONTAINS",
                          "values": ["email"]

                      }
                  ]
              }
          ]
      },
      {
          "name": "F4.contains.inverted",
          "description": "contains",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "email",
                          "operator": "STR_CONTAINS",
                          "values": ["email"],
                          "inverted": true

                      }
                  ]
              }
          ]
      },
      {
          "name": "F5.numEq",
          "description": "contains",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "someValue",
                          "operator": "NUM_EQ",
                          "value": "12"
                      }
                  ]
              }
          ]
      },
      {
          "name": "F5.numEq.float",
          "description": "contains",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "someValue",
                          "operator": "NUM_EQ",
                          "value": "12.0"
                      }
                  ]
              }
          ]
      },
      {
          "name": "F5.numEq.inverted",
          "description": "contains",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "someValue",
                          "operator": "NUM_EQ",
                          "value": "12",
                          "inverted": true
                      }
                  ]
              }
          ]
      },
      {
          "name": "F5.numGT",
          "description": "contains",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "someValue",
                          "operator": "NUM_GT",
                          "value": "12"
                      }
                  ]
              }
          ]
      },
      {
          "name": "F5.numGTE",
          "description": "contains",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "someValue",
                          "operator": "NUM_GTE",
                          "value": "12"
                      }
                  ]
              }
          ]
      },
      {
          "name": "F5.numLT",
          "description": "contains",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "someValue",
                          "operator": "NUM_LT",
                          "value": "12"
                      }
                  ]
              }
          ]
      },
      {
          "name": "F5.numLTE",
          "description": "contains",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "someValue",
                          "operator": "NUM_LTE",
                          "value": "12"
                      }
                  ]
              }
          ]
      },
      {
          "name": "F6.number-range",
          "description": "range of numbers",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "someValue",
                          "operator": "NUM_GT",
                          "value": "12"
                      },
                      {
                          "contextName": "someValue",
                          "operator": "NUM_LT",
                          "value": "16"
                      }
                  ]
              }
          ]
      },
      {
          "name": "F7.dateAfter",
          "description": "dates",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "currentTime",
                          "operator": "DATE_AFTER",
                          "value": "2022-01-29T13:00:00.000Z"
                      }
                  ]
              }
          ]
      },
      {
          "name": "F7.dateBefore",
          "description": "dates",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "currentTime",
                          "operator": "DATE_BEFORE",
                          "value": "2022-01-29T13:00:00.000Z"
                      }
                  ]
              }
          ]
      },
      {
          "name": "F7.date-range",
          "description": "dates",
          "enabled": true,
          "strategies": [
              {
                  "name": "default",
                  "parameters": {},
                  "constraints": [
                      {
                          "contextName": "currentTime",
                          "operator": "DATE_AFTER",
                          "value": "2022-01-22T13:00:00.000Z"
                      },
                      {
                          "contextName": "currentTime",
                          "operator": "DATE_BEFORE",
                          "value": "2022-01-29T13:00:00.000Z"
                      }
                  ]
              }
          ]
      }

  ]
  }
"""


@pytest.fixture()
def unleash_client():
    unleash_client = UnleashClient(url=URL,
                                   app_name=APP_NAME,
                                   instance_id='pytest_%s' % uuid.uuid4())
    yield unleash_client
    unleash_client.destroy()


@responses.activate
def test_feature_constraintoperators_startswith_enabled(unleash_client):
    """
    F1.startsWith should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F1.startsWith", {'customField': 'some-string-is-cool'})


@responses.activate
def test_feature_constraintoperators_startswith_disabled(unleash_client):
    """
    F1.startsWith should be disabled"
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F1.startsWith", {'customField': 'some2-string-is-cool'})


@responses.activate
def test_feature_constraintoperators_startswith_enabled_multiple(unleash_client):
    """
    F2.startsWith.multiple should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F2.startsWith.multiple", {'customField': 'e2 cool'})


@responses.activate
def test_feature_constraintoperators_startswith_disabled_multiple(unleash_client):
    """
    F2.startsWith.multiple should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F2.startsWith.multiple", {'customField': 'cool e2'})


@responses.activate
def test_feature_constraintoperators_endswith_enabled(unleash_client):
    """
    F3.endsWith should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F3.endsWith", {'email': '@some-email.com'})


@responses.activate
def test_feature_constraintoperators_endswith_incorrectcasing(unleash_client):
    """
    F3.endsWith should be disabled when casing is incorrect
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F3.endsWith", {'email': '@some-EMAIL.com'})


@responses.activate
def test_feature_constraintoperators_endswith_ignorecasing(unleash_client):
    """
    F3.endsWith should be disabled when casing is incorrect
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F3.endsWith.ignoringCase", {'email': '@SOME-EMAIL.com'})


@responses.activate
def test_feature_constraintoperators_endswith_disabled(unleash_client):
    """
    F3.endsWith should be disabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F3.endsWith", {'email': '@another-email.com'})


@responses.activate
def test_feature_constraintoperators_contains_enabled(unleash_client):
    """
    F4.contains should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F4.contains", {'email': '@some-email.com'})


@responses.activate
def test_feature_constraintoperators_contains_disabled(unleash_client):
    """
    F4.contains should be disabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F4.contains", {'email': '@another.com'})


@responses.activate
def test_feature_constraintoperators_contains_inverted(unleash_client):
    """
    F4.contains.inverted should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F4.contains.inverted", {'email': '@another.com'})


@responses.activate
def test_feature_constraintoperators_numeq_enabled(unleash_client):
    """
    F5.numEq
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F5.numEq", {'someValue': 12})


@responses.activate
def test_feature_constraintoperators_numeq_enabled_floats(unleash_client):
    """
    F5.numEq works for floats
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F5.numEq", {'someValue': 12.0})


@responses.activate
def test_feature_constraintoperators_numeq_inverted(unleash_client):
    """
    F5.numEq.inverted should be true
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F5.numEq.inverted", {'someValue': 12})


@responses.activate
def test_feature_constraintoperators_numeqfloat_enabled_floats(unleash_client):
    """
    F5.numEq.float works for floats
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F5.numEq.float", {'someValue': 12.0})


@responses.activate
def test_feature_constraintoperators_numeqfloat_enabled_int(unleash_client):
    """
    F5.numEq.float works for integers
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F5.numEq.float", {'someValue': 12})


@responses.activate
def test_feature_constraintoperators_numgt_enabled(unleash_client):
    """
    F5.numGT should be true
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F5.numGT", {'someValue': 13})


@responses.activate
def test_feature_constraintoperators_numgt_disabled(unleash_client):
    """
    F5.numGT should be false
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F5.numGT", {'someValue': 12})


@responses.activate
def test_feature_constraintoperators_numgte_enabled_equal(unleash_client):
    """
    F5.numGTE should be true when equal
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F5.numGTE", {'someValue': 12})


@responses.activate
def test_feature_constraintoperators_numgte_enabled_greater(unleash_client):
    """
    F5.numGTE should be true when larger
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F5.numGTE", {'someValue': 13})


@responses.activate
def test_feature_constraintoperators_numgte_disabled(unleash_client):
    """
    F5.numGTE should be false when lower
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F5.numGTE", {'someValue': 11})


@responses.activate
def test_feature_constraintoperators_numlte_enabled_equal(unleash_client):
    """
    F5.numLTE should be true when equal
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F5.numLTE", {'someValue': 12})


@responses.activate
def test_feature_constraintoperators_numlte_enabled_less(unleash_client):
    """
    F5.numLTE should be true when lower
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F5.numLTE", {'someValue': 0})


@responses.activate
def test_feature_constraintoperators_numlt_enabled_less(unleash_client):
    """
    F5.numLT should be true when lower
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F5.numLT", {'someValue': 0})


@responses.activate
def test_feature_constraintoperators_numlt_disabled_equal(unleash_client):
    """
    F5.numLT should be false when equal
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F5.numLT", {'someValue': 12})


@responses.activate
def test_feature_constraintoperators_numberranger_disabled(unleash_client):
    """
    F6.number-range should be false when not in range
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F6.number-range", {'someValue': 11})


@responses.activate
def test_feature_constraintoperators_numberranger_enabled(unleash_client):
    """
    F6.number-range should be true when in range
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F6.number-range", {'someValue': 14})


@responses.activate
def test_feature_constraintoperators_dateafter_enabled(unleash_client):
    """
    F7.dateAfter should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F7.dateAfter", {'currentTime': datetime(2022, 1, 30, 13)})


@responses.activate
def test_feature_constraintoperators_dateafter_disabled(unleash_client):
    """
    F7.dateAfter should be disabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F7.dateAfter", {'currentTime': datetime(2022, 1, 28, 13)})


@responses.activate
def test_feature_constraintoperators_dateafter_exclusive(unleash_client):
    """
    F7.dateAfter should be disabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F7.dateAfter", {'currentTime': datetime(2022, 1, 29, 13)})


@responses.activate
def test_feature_constraintoperators_datebefore_enabled(unleash_client):
    """
    F7.dateBefore should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F7.dateBefore", {'currentTime': datetime(2022, 1, 28, 13)})


@responses.activate
def test_feature_constraintoperators_datebefore_disabled(unleash_client):
    """
    F7.dateBefore should be disabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F7.dateBefore", {'currentTime': datetime(2022, 1, 30, 13)})


@responses.activate
def test_feature_constraintoperators_daterange(unleash_client):
    """
    "F7.data-range should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F7.date-range", {'currentTime': datetime(2022, 1, 25, 13)})
  


import pytest
import uuid
from datetime import datetime, timezone
from UnleashClient.constants import FEATURES_URL, METRIC_LAST_SENT_TIME, ETAG
from tests.utilities.mocks import MOCK_ALL_FEATURES, MOCK_CUSTOM_STRATEGY
from UnleashClient.cache import FileCache
from tests.utilities.mocks.mock_features import MOCK_FEATURES_WITH_SEGMENTS_RESPONSE


@pytest.fixture()
def cache_empty():
    cache_name = 'pytest_%s' % uuid.uuid4()
    temporary_cache = FileCache(cache_name)
    temporary_cache.mset({
        METRIC_LAST_SENT_TIME: datetime.now(timezone.utc),
        ETAG: ''
    })
    yield temporary_cache
    temporary_cache.destroy()


@pytest.fixture()
def cache_full():
    cache_name = 'pytest_%s' % uuid.uuid4()
    temporary_cache = FileCache(cache_name)
    temporary_cache.mset({
        FEATURES_URL: MOCK_ALL_FEATURES,
        METRIC_LAST_SENT_TIME: datetime.now(timezone.utc),
        ETAG: ''
    })
    yield temporary_cache
    temporary_cache.destroy()


@pytest.fixture()
def cache_custom():
    cache_name = 'pytest_%s' % uuid.uuid4()
    temporary_cache = FileCache(cache_name)
    temporary_cache.mset({
        FEATURES_URL: MOCK_CUSTOM_STRATEGY,
        METRIC_LAST_SENT_TIME: datetime.now(timezone.utc),
        ETAG: ''
    })
    yield temporary_cache
    temporary_cache.destroy()


@pytest.fixture()
def cache_segments():
    cache_name = 'pytest_%s' % uuid.uuid4()
    temporary_cache = FileCache(cache_name)
    temporary_cache.mset({
        FEATURES_URL: MOCK_FEATURES_WITH_SEGMENTS_RESPONSE,
        METRIC_LAST_SENT_TIME: datetime.now(timezone.utc),
        ETAG: ''
    })
    yield temporary_cache
    temporary_cache.destroy()


# pylint: disable=invalid-name
import random
from UnleashClient.strategies.Strategy import Strategy


class GradualRolloutRandom(Strategy):
    def apply(self, context: dict = None) -> bool:
        """
        Returns random assignment.

        :return:
        """
        percentage = int(self.parameters["percentage"])

        return percentage > 0 and random.randint(1, 100) <= percentage



import warnings
from UnleashClient.strategies import Strategy


def strategy_v2xx_deprecation_check(strategies: list) -> None:
    """
    Notify users of backwards incompatible changes in v3 for custom strategies.
    """
    for strategy in strategies:
        try:
            # Check if the __call__() method is overwritten (should only be true for custom strategies in v1.x or v2.x.
            if strategy.__call__ != Strategy.__call__:
                warnings.warn(
                    f"unleash-client-python v3.x.x requires overriding the execute() method instead of the __call__() method. Error in: {strategy.__name__}",
                    DeprecationWarning
                )
        except AttributeError:
            # Ignore if not.
            pass



import pytest
from UnleashClient.variants import Variants
from tests.utilities.mocks.mock_variants import VARIANTS, VARIANTS_WITH_STICKINESS


@pytest.fixture()
def variations():
    yield Variants(VARIANTS, "TestFeature")


@pytest.fixture()
def variations_with_stickiness():
    yield Variants(VARIANTS_WITH_STICKINESS, "TestFeature")


def test_variations_override_match(variations):
    override_variant = variations._apply_overrides({'userId': '1'})
    assert override_variant['name'] == 'VarA'


def test_variations_overrid_nomatch(variations):
    assert not variations._apply_overrides({'userId': '2'})


def test_variations_seed(variations):
    # Random seed generation
    context = {}
    seed = variations._get_seed(context)
    assert float(seed) > 0

    # UserId, SessionId, and remoteAddress
    context = {
        'userId': '1',
        'sessionId': '1',
        'remoteAddress': '1.1.1.1'
    }

    assert context['userId'] == variations._get_seed(context)
    del context['userId']
    assert context['sessionId'] == variations._get_seed(context)
    del context['sessionId']
    assert context['remoteAddress'] == variations._get_seed(context)


def test_variations_seed_override(variations):
    # UserId, SessionId, and remoteAddress
    context = {
        'userId': '1',
        'sessionId': '1',
        'remoteAddress': '1.1.1.1',
        'customField': "ActuallyAmAHamster"
    }

    assert context['customField'] == variations._get_seed(context, 'customField')


def test_variation_selectvariation_happypath(variations):
    variant = variations.get_variant({'userId': '2'})
    assert variant
    assert 'payload' in variant
    assert variant['name'] == 'VarC'


def test_variation_customvariation(variations_with_stickiness):
    variations = variations_with_stickiness
    variant = variations.get_variant({'customField': 'ActuallyAmAHamster1234'})
    assert variant
    assert 'payload' in variant
    assert variant['name'] == 'VarC'


def test_variation_selectvariation_multi(variations):
    tracker = {}
    for x in range(100):
        variant = variations.get_variant({})
        name = variant['name']
        if name in tracker:
            tracker[name] += 1
        else:
            tracker[name] = 1

    assert len(tracker) == 3
    assert sum([tracker[x] for x in tracker.keys()]) == 100


def test_variation_override(variations):
    variant = variations.get_variant({'userId': '1'})
    assert variant
    assert 'payload' in variant
    assert variant['name'] == 'VarA'


def test_variation_novariants():
    variations = Variants([], "TestFeature")
    variant = variations.get_variant({})
    assert variant
    assert variant['name'] == 'disabled'



import uuid
import json
import pytest
import responses
from UnleashClient import UnleashClient
from UnleashClient.constants import FEATURES_URL, METRICS_URL, REGISTER_URL
from tests.utilities.testing_constants import URL, APP_NAME


MOCK_JSON = """
{
  "version": 2,
  "features": [
    {
      "name": "F9.globalSegmentOn",
      "description": "With global segment referencing constraint in on state",
      "enabled": true,
      "strategies": [
        {
          "name": "default",
          "parameters": {},
          "segments": [1]
        }
      ]
    },
    {
      "name": "F9.globalSegmentOff",
      "description": "With global segment referencing constraint in off state",
      "enabled": true,
      "strategies": [
        {
          "name": "default",
          "parameters": {},
          "segments": [2]
        }
      ]
    },
    {
      "name": "F9.globalSegmentAndConstraint",
      "description": "With global segment and constraint both on",
      "enabled": true,
      "strategies": [
        {
          "name": "default",
          "parameters": {},
          "constraints": [
            {
              "contextName": "version",
              "operator": "SEMVER_EQ",
              "value": "1.2.2"
            }
          ],
          "segments": [1]
        }
      ]
    },
    {
      "name": "F9.withMissingSegment",
      "description": "With global segment that doesn't exist",
      "enabled": true,
      "strategies": [
        {
          "name": "default",
          "parameters": {},
          "constraints": [
            {
              "contextName": "version",
              "operator": "SEMVER_EQ",
              "value": "1.2.2"
            }
          ],
          "segments": [3]
        }
      ]
    },
    {
      "name": "F9.withSeveralConstraintsAndSegments",
      "description": "With several segments and constraints",
      "enabled": true,
      "strategies": [
        {
          "name": "default",
          "parameters": {},
          "constraints": [
            {
              "contextName": "customNumber",
              "operator": "NUM_LT",
              "value": "10"
            },
            {
              "contextName": "version",
              "operator": "SEMVER_LT",
              "value": "3.2.2"
            }
          ],
          "segments": [1, 4, 5]
        }
      ]
    }
  ],
  "segments": [
    {
      "id": 1,
      "constraints": [
        {
          "contextName": "version",
          "operator": "SEMVER_EQ",
          "value": "1.2.2"
        }
      ]
    },
    {
      "id": 2,
      "constraints": [
        {
          "contextName": "version",
          "operator": "SEMVER_EQ",
          "value": "3.1.4"
        }
      ]
    },
    {
      "id": 3,
      "constraints": [
        {
          "contextName": "version",
          "operator": "SEMVER_EQ",
          "value": "3.1.4"
        }
      ]
    },
    {
      "id": 4,
      "constraints": [
        {
          "contextName": "customName",
          "operator": "STR_CONTAINS",
          "values": ["Pi"]
        }
      ]
    },
    {
      "id": 5,
      "constraints": [
        {
          "contextName": "slicesLeft",
          "operator": "NUM_LTE",
          "value": "4"
        }
      ]
    }
  ]
}
"""


@pytest.fixture()
def unleash_client():
    unleash_client = UnleashClient(url=URL,
                                   app_name=APP_NAME,
                                   instance_id='pytest_%s' % uuid.uuid4())
    yield unleash_client
    unleash_client.destroy()


@responses.activate
def test_feature_globalSegmentOff_should_be_disabled(unleash_client):
    """
    F9.globalSegmentOff should be disabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F9.globalSegmentOff", {'version': '1.2.2'})


@responses.activate
def test_feature_globalSegmentOn_should_be_enabled(unleash_client):
    """
    F9.globalSegmentOn should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F9.globalSegmentOn", {'version': '1.2.2'})


@responses.activate
def test_feature_globalSegmentAndConstraint_should_be_enabled(unleash_client):
    """
    F9.globalSegmentAndConstraint should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("F9.globalSegmentAndConstraint", {'version': '1.2.2'})


@responses.activate
def test_feature_withMissingSegment_should_force_evaluation_to_false(unleash_client):
    """
    F9.withMissingSegment should force evaluation to false
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F9.withMissingSegment", {'version': '1.2.2'})


@responses.activate
def test_feature_withSeveralConstraintsAndSegments_should_be_enabled(unleash_client):
    """
    F9.withSeveralConstraintsAndSegments should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("F9.withSeveralConstraintsAndSegments", {"version": "1.2.2",
                                                                                  "customNumber": "3.14",
                                                                                  "customName": "Pie",
                                                                                  "slicesLeft": "4"})



import responses
from UnleashClient.constants import FEATURES_URL, ETAG
from UnleashClient.periodic_tasks import fetch_and_load_features
from UnleashClient.features import Feature
from tests.utilities.mocks.mock_features import MOCK_FEATURE_RESPONSE, MOCK_FEATURE_RESPONSE_PROJECT
from tests.utilities.testing_constants import URL, APP_NAME, INSTANCE_ID, CUSTOM_HEADERS, CUSTOM_OPTIONS, DEFAULT_STRATEGY_MAPPING, PROJECT_URL, PROJECT_NAME, ETAG_VALUE


FULL_FEATURE_URL = URL + FEATURES_URL


@responses.activate
def test_fetch_and_load(cache_empty):  # noqa: F811
    # Set up for tests
    in_memory_features = {}
    responses.add(responses.GET, FULL_FEATURE_URL, json=MOCK_FEATURE_RESPONSE, status=200, headers={'etag': ETAG_VALUE})
    temp_cache = cache_empty

    fetch_and_load_features(URL,
                            APP_NAME,
                            INSTANCE_ID,
                            CUSTOM_HEADERS,
                            CUSTOM_OPTIONS,
                            temp_cache,
                            in_memory_features,
                            DEFAULT_STRATEGY_MAPPING)

    assert isinstance(in_memory_features["testFlag"], Feature)
    assert temp_cache.get(ETAG) == ETAG_VALUE


@responses.activate
def test_fetch_and_load_project(cache_empty):  # noqa: F811
    # Set up for tests
    in_memory_features = {}
    responses.add(responses.GET, PROJECT_URL, json=MOCK_FEATURE_RESPONSE_PROJECT, status=200)
    temp_cache = cache_empty

    fetch_and_load_features(URL,
                            APP_NAME,
                            INSTANCE_ID,
                            CUSTOM_HEADERS,
                            CUSTOM_OPTIONS,
                            temp_cache,
                            in_memory_features,
                            DEFAULT_STRATEGY_MAPPING,
                            PROJECT_NAME)

    assert len(in_memory_features.keys()) == 1
    assert isinstance(in_memory_features["ivan-project"], Feature)


@responses.activate
def test_fetch_and_load_failure(cache_empty):  # noqa: F811
    # Set up for tests
    in_memory_features = {}
    responses.add(responses.GET, FULL_FEATURE_URL, json=MOCK_FEATURE_RESPONSE, status=200)
    temp_cache = cache_empty

    fetch_and_load_features(URL,
                            APP_NAME,
                            INSTANCE_ID,
                            CUSTOM_HEADERS,
                            CUSTOM_OPTIONS,
                            temp_cache,
                            in_memory_features,
                            DEFAULT_STRATEGY_MAPPING)

    # Fail next request
    responses.reset()
    responses.add(responses.GET, FULL_FEATURE_URL, json={}, status=500)

    fetch_and_load_features(URL,
                            APP_NAME,
                            INSTANCE_ID,
                            CUSTOM_HEADERS,
                            CUSTOM_OPTIONS,
                            temp_cache,
                            in_memory_features,
                            DEFAULT_STRATEGY_MAPPING)

    assert isinstance(in_memory_features["testFlag"], Feature)



# pylint: disable=invalid-name, too-few-public-methods, use-a-generator
from typing import Optional, Union
from datetime import datetime
from enum import Enum
from dateutil.parser import parse, ParserError
import semver
from UnleashClient.utils import LOGGER, get_identifier


class ConstraintOperators(Enum):
    # Logical operators
    IN = "IN"
    NOT_IN = "NOT_IN"

    # String operators
    STR_ENDS_WITH = "STR_ENDS_WITH"
    STR_STARTS_WITH = "STR_STARTS_WITH"
    STR_CONTAINS = "STR_CONTAINS"

    # Numeric oeprators
    NUM_EQ = "NUM_EQ"
    NUM_GT = "NUM_GT"
    NUM_GTE = "NUM_GTE"
    NUM_LT = "NUM_LT"
    NUM_LTE = "NUM_LTE"

    # Date operators
    DATE_AFTER = "DATE_AFTER"
    DATE_BEFORE = "DATE_BEFORE"

    # Semver operators
    SEMVER_EQ = "SEMVER_EQ"
    SEMVER_GT = "SEMVER_GT"
    SEMVER_LT = "SEMVER_LT"


class Constraint:
    def __init__(self, constraint_dict: dict) -> None:
        """
        Represents a constraint on a strategy

        :param constraint_dict: From the strategy document.
        """
        self.context_name: str = constraint_dict['contextName']
        self.operator: ConstraintOperators = ConstraintOperators(constraint_dict['operator'].upper())
        self.values = constraint_dict['values'] if 'values' in constraint_dict.keys() else []
        self.value = constraint_dict['value'] if 'value' in constraint_dict.keys() else None

        self.case_insensitive = constraint_dict['caseInsensitive'] if 'caseInsensitive' in constraint_dict.keys() else False
        self.inverted = constraint_dict['inverted'] if 'inverted' in constraint_dict.keys() else False


    # Methods to handle each operator type.
    def check_list_operators(self, context_value: str) -> bool:
        return_value = False

        if self.operator == ConstraintOperators.IN:
            return_value = context_value in self.values
        elif self.operator == ConstraintOperators.NOT_IN:
            return_value = context_value not in self.values

        return return_value

    def check_string_operators(self, context_value: str) -> bool:
        if self.case_insensitive:
            normalized_values = [x.upper() for x in self.values]
            normalized_context_value = context_value.upper()
        else:
            normalized_values = self.values
            normalized_context_value = context_value

        return_value = False

        if self.operator == ConstraintOperators.STR_CONTAINS:
            return_value = any([x in normalized_context_value for x in normalized_values])
        elif self.operator == ConstraintOperators.STR_ENDS_WITH:
            return_value = any([normalized_context_value.endswith(x) for x in normalized_values])
        elif self.operator == ConstraintOperators.STR_STARTS_WITH:
            return_value = any([normalized_context_value.startswith(x) for x in normalized_values])

        return return_value

    def check_numeric_operators(self, context_value: Union[float, int]) -> bool:
        return_value = False
        parsed_value = float(self.value)

        if self.operator == ConstraintOperators.NUM_EQ:
            return_value = context_value == parsed_value
        elif self.operator == ConstraintOperators.NUM_GT:
            return_value = context_value > parsed_value
        elif self.operator == ConstraintOperators.NUM_GTE:
            return_value = context_value >= parsed_value
        elif self.operator == ConstraintOperators.NUM_LT:
            return_value = context_value < parsed_value
        elif self.operator == ConstraintOperators.NUM_LTE:
            return_value = context_value <= parsed_value

        return return_value


    def check_date_operators(self, context_value: datetime) -> bool:
        return_value = False
        parsing_exception = False

        try:
            parsed_date = parse(self.value, ignoretz=True)
        except ParserError:
            LOGGER.error(f"Unable to parse date: {self.value}")
            parsing_exception = True

        if not parsing_exception:
            if self.operator == ConstraintOperators.DATE_AFTER:
                return_value = context_value > parsed_date
            elif self.operator == ConstraintOperators.DATE_BEFORE:
                return_value = context_value < parsed_date

        return return_value


    def check_semver_operators(self, context_value: str) -> bool:
        return_value = False
        parsing_exception = False
        target_version: Optional[semver.VersionInfo] = None
        context_version: Optional[semver.VersionInfo] = None

        try:
            target_version = semver.VersionInfo.parse(self.value)
        except ValueError:
            LOGGER.error(f"Unable to parse server semver: {self.value}")
            parsing_exception = True

        try:
            context_version = semver.VersionInfo.parse(context_value)
        except ValueError:
            LOGGER.error(f"Unable to parse context semver: {context_value}")
            parsing_exception = True

        if not parsing_exception:
            if self.operator == ConstraintOperators.SEMVER_EQ:
                return_value = context_version == target_version
            elif self.operator == ConstraintOperators.SEMVER_GT:
                return_value = context_version > target_version
            elif self.operator == ConstraintOperators.SEMVER_LT:
                return_value = context_version < target_version

        return return_value


    def apply(self, context: dict = None) -> bool:
        """
        Returns true/false depending on constraint provisioning and context.

        :param context: Context information
        :return:
        """
        constraint_check = False

        try:
            context_value = get_identifier(self.context_name, context)

            # Set currentTime if not specified
            if self.context_name == "currentTime" and not context_value:
                context_value = datetime.now()

            if context_value is not None:
                if self.operator in [ConstraintOperators.IN, ConstraintOperators.NOT_IN]:
                    constraint_check = self.check_list_operators(context_value=context_value)
                elif self.operator in [ConstraintOperators.STR_CONTAINS, ConstraintOperators.STR_ENDS_WITH, ConstraintOperators.STR_STARTS_WITH]:
                    constraint_check = self.check_string_operators(context_value=context_value)
                elif self.operator in [ConstraintOperators.NUM_EQ, ConstraintOperators.NUM_GT, ConstraintOperators.NUM_GTE, ConstraintOperators.NUM_LT, ConstraintOperators.NUM_LTE]:
                    constraint_check = self.check_numeric_operators(context_value=context_value)
                elif self.operator in [ConstraintOperators.DATE_AFTER, ConstraintOperators.DATE_BEFORE]:
                    constraint_check = self.check_date_operators(context_value=context_value)
                elif self.operator in [ConstraintOperators.SEMVER_EQ, ConstraintOperators.SEMVER_GT, ConstraintOperators.SEMVER_LT]:
                    constraint_check = self.check_semver_operators(context_value=context_value)

        except Exception as excep:  # pylint: disable=broad-except
            LOGGER.info("Could not evaluate context %s!  Error: %s", self.context_name, excep)

        return not constraint_check if self.inverted else constraint_check



import json
from datetime import datetime, timezone, timedelta
import responses
from tests.utilities.testing_constants import URL, APP_NAME, INSTANCE_ID, CUSTOM_HEADERS, CUSTOM_OPTIONS, IP_LIST
from UnleashClient.constants import METRICS_URL, METRIC_LAST_SENT_TIME
from UnleashClient.periodic_tasks import aggregate_and_send_metrics
from UnleashClient.features import Feature
from UnleashClient.strategies import RemoteAddress, Default
from UnleashClient.cache import FileCache


FULL_METRICS_URL = URL + METRICS_URL
print(FULL_METRICS_URL)


@responses.activate
def test_aggregate_and_send_metrics():
    responses.add(responses.POST, FULL_METRICS_URL, json={}, status=200)

    start_time = datetime.now(timezone.utc) - timedelta(seconds=60)
    cache = FileCache("TestCache")
    cache.set(METRIC_LAST_SENT_TIME, start_time)
    strategies = [RemoteAddress(parameters={"IPs": IP_LIST}), Default()]
    my_feature1 = Feature("My Feature1", True, strategies)
    my_feature1.yes_count = 1
    my_feature1.no_count = 1

    my_feature2 = Feature("My Feature2", True, strategies)
    my_feature2.yes_count = 2
    my_feature2.no_count = 2

    my_feature3 = Feature("My Feature3", True, strategies)
    my_feature3.yes_count = 0
    my_feature3.no_count = 0

    features = {"My Feature1": my_feature1, "My Feature 2": my_feature2}

    aggregate_and_send_metrics(URL, APP_NAME, INSTANCE_ID, CUSTOM_HEADERS, CUSTOM_OPTIONS, features, cache)

    assert len(responses.calls) == 1
    request = json.loads(responses.calls[0].request.body)

    assert len(request['bucket']["toggles"].keys()) == 2
    assert request['bucket']["toggles"]["My Feature1"]["yes"] == 1
    assert request['bucket']["toggles"]["My Feature1"]["no"] == 1
    assert "My Feature3" not in request['bucket']["toggles"].keys()
    assert cache.get(METRIC_LAST_SENT_TIME) > start_time


@responses.activate
def test_no_metrics():
    responses.add(responses.POST, FULL_METRICS_URL, json={}, status=200)

    start_time = datetime.now(timezone.utc) - timedelta(seconds=60)
    cache = FileCache("TestCache")
    cache.set(METRIC_LAST_SENT_TIME, start_time)
    strategies = [RemoteAddress(parameters={"IPs": IP_LIST}), Default()]

    my_feature1 = Feature("My Feature1", True, strategies)
    my_feature1.yes_count = 0
    my_feature1.no_count = 0

    features = {"My Feature1": my_feature1}

    aggregate_and_send_metrics(URL, APP_NAME, INSTANCE_ID, CUSTOM_HEADERS, CUSTOM_OPTIONS, features, cache)

    assert len(responses.calls) == 0



from datetime import datetime
from sqlite3 import Date
import pytest
from UnleashClient.constraints import Constraint
from tests.utilities.mocks import mock_constraints


@pytest.fixture()
def constraint_IN():
    yield Constraint(mock_constraints.CONSTRAINT_DICT_IN)


@pytest.fixture()
def constraint_NOTIN():
    yield Constraint(mock_constraints.CONSTRAINT_DICT_NOTIN)



def test_constraint_IN_match(constraint_IN):
    constraint = constraint_IN
    context = {
        'appName': 'test'
    }

    assert constraint.apply(context)


def test_constraint_IN_not_match(constraint_IN):
    constraint = constraint_IN
    context = {
        'appName': 'test3'
    }

    assert not constraint.apply(context)


def test_constraint_IN_missingcontext(constraint_IN):
    constraint = constraint_IN
    assert not constraint.apply({})


def test_constraint_NOTIN_match(constraint_NOTIN):
    constraint = constraint_NOTIN
    context = {
        'appName': 'test'
    }

    assert not constraint.apply(context)


def test_constraint_NOTIN_not_match(constraint_NOTIN):
    constraint = constraint_NOTIN
    context = {
        'appName': 'test3'
    }

    assert constraint.apply(context)


def test_constraint_inversion():
    constraint_ci = Constraint(constraint_dict=mock_constraints.CONSTRAINT_DICT_STR_INVERT)

    assert not constraint_ci.apply({'customField': "adogb"})


def test_constraint_STR_CONTAINS():
    constraint_not_ci = Constraint(constraint_dict=mock_constraints.CONSTRAINT_DICT_STR_CONTAINS_NOT_CI)
    constraint_ci = Constraint(constraint_dict=mock_constraints.CONSTRAINT_DICT_STR_CONTAINS_CI)
    
    assert constraint_ci.apply({'customField': "adogb"})
    assert not constraint_ci.apply({'customField': "aparrotb"})
    assert constraint_ci.apply({'customField': "ahamsterb"})

    assert constraint_not_ci.apply({'customField': "adogb"})
    assert not constraint_ci.apply({'customField': "aparrotb"})
    assert not constraint_not_ci.apply({'customField': "ahamsterb"})


def test_constraint_STR_ENDS_WITH():
    constraint_not_ci = Constraint(constraint_dict=mock_constraints.CONSTRAINT_DICT_STR_ENDS_WITH_NOT_CI)
    constraint_ci = Constraint(constraint_dict=mock_constraints.CONSTRAINT_DICT_STR_ENDS_WITH_CI)
    
    assert constraint_ci.apply({'customField': "adog"})
    assert not constraint_ci.apply({'customField': "aparrot"})
    assert constraint_ci.apply({'customField': "ahamster"})

    assert constraint_not_ci.apply({'customField': "adog"})
    assert not constraint_not_ci.apply({'customField': "aparrot"})
    assert not constraint_not_ci.apply({'customField': "ahamster"})


def test_constraint_STR_STARTS_WITH():
    constraint_not_ci = Constraint(constraint_dict=mock_constraints.CONSTRAINT_DICT_STR_STARTS_WITH_NOT_CI)
    constraint_ci = Constraint(constraint_dict=mock_constraints.CONSTRAINT_DICT_STR_STARTS_WITH_CI)
    
    assert constraint_ci.apply({'customField': "dogb"})
    assert not constraint_ci.apply({'customField': "parrotb"})
    assert constraint_ci.apply({'customField': "hamsterb"})

    assert constraint_not_ci.apply({'customField': "dogb"})
    assert not constraint_not_ci.apply({'customField': "parrotb"})
    assert not constraint_not_ci.apply({'customField': "hamsterb"})


def test_constraints_NUM_EQ():
    constraint = Constraint(constraint_dict=mock_constraints.CONSTRAINT_NUM_EQ)

    assert not constraint.apply({'customField': 4})
    assert constraint.apply({'customField': 5})
    assert not constraint.apply({'customField': 6})


def test_constraints_NUM_GT():
    constraint = Constraint(constraint_dict=mock_constraints.CONSTRAINT_NUM_GT)

    assert not constraint.apply({'customField': 4})
    assert not constraint.apply({'customField': 5})
    assert constraint.apply({'customField': 6})


def test_constraints_NUM_GTE():
    constraint = Constraint(constraint_dict=mock_constraints.CONSTRAINT_NUM_GTE)

    assert not constraint.apply({'customField': 4})
    assert constraint.apply({'customField': 5})
    assert constraint.apply({'customField': 6})


def test_constraints_NUM_LT():
    constraint = Constraint(constraint_dict=mock_constraints.CONSTRAINT_NUM_LT)

    assert constraint.apply({'customField': 4})
    assert not constraint.apply({'customField': 5})
    assert not constraint.apply({'customField': 6})


def test_constraints_NUM_LTE():
    constraint = Constraint(constraint_dict=mock_constraints.CONSTRAINT_NUM_LTE)

    assert constraint.apply({'customField': 4})
    assert constraint.apply({'customField': 5})
    assert not constraint.apply({'customField': 6})


def test_constraints_NUM_FLOAT():
    constraint = Constraint(constraint_dict=mock_constraints.CONSTRAINT_NUM_FLOAT)

    assert constraint.apply({'customField': 5})
    assert constraint.apply({'customField': 5.1})
    assert not constraint.apply({'customField': 5.2})


def test_constraints_DATE_AFTER():
    constraint = Constraint(constraint_dict=mock_constraints.CONSTRAINT_DATE_AFTER)

    assert constraint.apply({'currentTime': datetime(2022, 1, 23)})
    assert not constraint.apply({'currentTime': datetime(2022, 1, 22)})
    assert not constraint.apply({'currentTime': datetime(2022, 1, 21)})


def test_constraints_DATE_BEFORE():
    constraint = Constraint(constraint_dict=mock_constraints.CONSTRAINT_DATE_BEFORE)

    assert not constraint.apply({'currentTime': datetime(2022, 1, 23)})
    assert not constraint.apply({'currentTime': datetime(2022, 1, 22)})
    assert constraint.apply({'currentTime': datetime(2022, 1, 21)})


def test_constraints_default():
    constraint = Constraint(constraint_dict=mock_constraints.CONSTRAINT_DATE_BEFORE)

    assert not constraint.apply({})


def test_constraints_date_error():
    constraint = Constraint(constraint_dict=mock_constraints.CONSTRAINT_DATE_ERROR)
    assert not constraint.apply({'currentTime': datetime(2022, 1, 23)})


def test_constraints_SEMVER_EQ():
    constraint = Constraint(constraint_dict=mock_constraints.CONSTRAINT_SEMVER_EQ)

    assert not constraint.apply({'customField': '1.2.1'})
    assert constraint.apply({'customField': '1.2.2'})
    assert not constraint.apply({'customField': '1.2.3'})


def test_constraints_SEMVER_GT():
    constraint = Constraint(constraint_dict=mock_constraints.CONSTRAINT_SEMVER_GT)

    assert not constraint.apply({'customField': '1.2.1'})
    assert not constraint.apply({'customField': '1.2.2'})
    assert constraint.apply({'customField': '1.2.3'})


def test_constraints_SEMVER_LT():
    constraint = Constraint(constraint_dict=mock_constraints.CONSTRAINT_SEMVER_LT)

    assert constraint.apply({'customField': '1.2.1'})
    assert not constraint.apply({'customField': '1.2.2'})
    assert not constraint.apply({'customField': '1.2.3'})


def test_constraints_semverexception():
    constraint = Constraint(constraint_dict=mock_constraints.CONSTRAINT_SEMVER_EQ)

    assert not constraint.apply({'customField': 'hamstershamsterhamsters'})



import pytest
from UnleashClient.strategies import GradualRolloutUserId
from tests.utilities import generate_context


@pytest.fixture()
def strategy():
    yield GradualRolloutUserId(parameters={"percentage": 50, "groupId": "test"})


def test_userwithid(strategy):
    strategy.execute(context=generate_context())



from importlib_metadata import version

# Library
SDK_NAME = "unleash-client-python"
SDK_VERSION = version("UnleashClient")
REQUEST_TIMEOUT = 30
REQUEST_RETRIES = 3
METRIC_LAST_SENT_TIME = "mlst"

# =Unleash=
APPLICATION_HEADERS = {"Content-Type": "application/json", "Unleash-Client-Spec": "4.2.0"}
DISABLED_VARIATION = {
    'name': 'disabled',
    'enabled': False
}

# Paths
REGISTER_URL = "/client/register"
FEATURES_URL = "/client/features"
METRICS_URL = "/client/metrics"

# Cache keys
FAILED_STRATEGIES = 'failed_strategies'
ETAG = 'etag'



import logging
from typing import Any
import mmh3  # pylint: disable=import-error
from requests import Response

LOGGER = logging.getLogger('UnleashClient')

def normalized_hash(identifier: str,
                    activation_group: str,
                    normalizer: int = 100) -> int:
    return mmh3.hash(f"{activation_group}:{identifier}", signed=False) % normalizer + 1


def get_identifier(context_key_name: str, context: dict) -> Any:
    if context_key_name in context.keys():
        value = context[context_key_name]
    elif 'properties' in context.keys() and context_key_name in context['properties'].keys():
        value = context['properties'][context_key_name]
    else:
        value = None

    return value


def log_resp_info(resp: Response) -> None:
    LOGGER.debug("HTTP status code: %s", resp.status_code)
    LOGGER.debug("HTTP headers: %s", resp.headers)
    LOGGER.debug("HTTP content: %s", resp.text)



import json
from datetime import datetime, timezone
import requests
from requests.exceptions import MissingSchema, InvalidSchema, InvalidURL, InvalidHeader
from UnleashClient.constants import SDK_NAME, SDK_VERSION, REQUEST_TIMEOUT, APPLICATION_HEADERS, REGISTER_URL
from UnleashClient.utils import LOGGER, log_resp_info


# pylint: disable=broad-except
def register_client(url: str,
                    app_name: str,
                    instance_id: str,
                    metrics_interval: int,
                    custom_headers: dict,
                    custom_options: dict,
                    supported_strategies: dict) -> bool:
    """
    Attempts to register client with unleash server.

    Notes:
    * If unsuccessful (i.e. not HTTP status code 202), exception will be caught and logged.
      This is to allow "safe" error handling if unleash server goes down.

    :param url:
    :param app_name:
    :param instance_id:
    :param metrics_interval:
    :param custom_headers:
    :param custom_options:
    :param supported_strategies:
    :return: true if registration successful, false if registration unsuccessful or exception.
    """
    registation_request = {
        "appName": app_name,
        "instanceId": instance_id,
        "sdkVersion": f"{SDK_NAME}:{SDK_VERSION}",
        "strategies": [*supported_strategies],
        "started": datetime.now(timezone.utc).isoformat(),
        "interval": metrics_interval
    }

    try:
        LOGGER.info("Registering unleash client with unleash @ %s", url)
        LOGGER.info("Registration request information: %s", registation_request)

        resp = requests.post(url + REGISTER_URL,
                             data=json.dumps(registation_request),
                             headers={**custom_headers, **APPLICATION_HEADERS},
                             timeout=REQUEST_TIMEOUT, **custom_options)

        if resp.status_code != 202:
            log_resp_info(resp)
            LOGGER.warning("Unleash Client registration failed due to unexpected HTTP status code.")
            return False

        LOGGER.info("Unleash Client successfully registered!")

        return True
    except (MissingSchema, InvalidSchema, InvalidHeader, InvalidURL) as exc:
        LOGGER.exception("Unleash Client registration failed fatally due to exception: %s", exc)
        raise exc
    except requests.RequestException as exc:
        LOGGER.exception("Unleash Client registration failed due to exception: %s", exc)

    return False



# pylint: disable=invalid-name
from UnleashClient.strategies.Strategy import Strategy


class UserWithId(Strategy):
    def load_provisioning(self) -> list:
        return [x.strip() for x in self.parameters["userIds"].split(',')]

    def apply(self, context: dict = None) -> bool:
        """
        Returns true if userId is a member of id list.

        :return:
        """
        return_value = False

        if "userId" in context.keys():
            return_value = context["userId"] in self.parsed_provisioning

        return return_value



MOCK_FEATURE_RESPONSE = {
    "version": 1,
    "features": [
        {
            "name": "testFlag",
            "description": "This is a test!",
            "enabled": True,
            "strategies": [
                {
                    "name": "default",
                    "parameters": {}
                }
            ],
            "createdAt": "2018-10-04T01:27:28.477Z"
        },
        {
            "name": "testFlag2",
            "description": "Test flag 2",
            "enabled": True,
            "strategies": [
                {
                    "name": "gradualRolloutRandom",
                    "parameters": {
                        "percentage": 50
                    }
                }
            ],
            "createdAt": "2018-10-04T11:03:56.062Z"
        },
        {
            "name": "testContextFlag",
            "description": "This is a test for static context fileds!",
            "enabled": True,
            "strategies": [
                {
                    "name": "custom-context",
                    "parameters": {
                        "environments": "prod"
                    }
                }
            ],
            "createdAt": "2018-10-04T01:27:28.477Z"
        },
        {
            "name": "testConstraintFlag",
            "description": "This is a flag with a constraint!",
            "enabled": True,
            "strategies": [
                {
                    "name": "default",
                    "parameters": {},
                    "constraints": [
                        {
                            "contextName": "currentTime",
                            "operator": "DATE_BEFORE",
                            "value": "2022-01-22T00:00:00.000Z",
                            "inverted": False
                        }
                    ],
                },
            ],
            "createdAt": "2018-10-04T01:27:28.477Z"
        },
        {
          "name": "testVariations",
          "description": "Test variation",
          "enabled": True,
          "strategies": [
              {
                  "name": "userWithId",
                  "parameters": {
                      "userIds": "2"
                  }
              }
          ],
          "variants": [
              {
                  "name": "VarA",
                  "weight": 34,
                  "payload": {
                      "type": "string",
                      "value": "Test1"
                  },
                  "overrides": [
                      {
                          "contextName": "userId",
                          "values": [
                              "ivanklee86@gmail.com",
                              "ivan@aaptiv.com"
                          ]
                      }
                  ]
              },
              {
                  "name": "VarB",
                  "weight": 33,
                  "payload": {
                      "type": "string",
                      "value": "Test 2"
                  }
              },
              {
                  "name": "VarC",
                  "weight": 33,
                  "payload": {
                      "type": "string",
                      "value": "Test 3"
                  }
              }
          ],
          "createdAt": "2019-10-25T13:22:02.035Z"
      }
    ]
}

MOCK_FEATURES_WITH_SEGMENTS_RESPONSE = {
    "version": 2,
    "features": [
        {
            "strategies": [
                {
                    "name": "default",
                    "constraints": [],
                    "parameters": {},
                    "segments": [
                        1, 2
                    ]
                }
            ],
            "impressionData": False,
            "enabled": True,
            "name": "Test",
            "description": "",
            "project": "default",
            "stale": False,
            "type": "release",
            "variants": []
        }
    ],
    "query": {
        "environment": "development",
        "inlineSegmentConstraints": False
    },
    "segments": [
        {
            "id": 1,
            "name": "TestSegment1",
            "description": "test",
            "constraints": [
                {
                    "value": "2022-06-14T06:40:17.766Z",
                    "values": [],
                    "inverted": False,
                    "operator": "DATE_BEFORE",
                    "contextName": "currentTime",
                    "caseInsensitive": False
                }
            ],
            "createdBy": "admin",
            "createdAt": "2022-06-14T06:40:25.331Z"
        },
        {
            "id": 2,
            "name": "TestSegment2",
            "description": "test",
            "constraints": [
                {
                    "value": "2022-06-14T06:40:17.766Z",
                    "values": [],
                    "inverted": False,
                    "operator": "DATE_AFTER",
                    "contextName": "currentTime",
                    "caseInsensitive": False
                }
            ],
            "createdBy": "admin",
            "createdAt": "2022-06-14T06:40:25.331Z"
        }
    ]
}

MOCK_FEATURE_RESPONSE_PROJECT = {
   "version":1,
   "features":[
      {
         "name":"ivan-project",
         "type":"release",
         "enabled": True,
         "stale": False,
         "strategies":[
            {
               "name":"default",
               "parameters":{

               }
            }
         ],
         "variants":[

         ]
      }
   ]
}


MOCK_CUSTOM_STRATEGY = {
  "version": 1,
  "features": [
    {
      "name": "CustomToggle",
      "description": "CustomToggle Test",
      "enabled": True,
      "strategies": [
        {
          "name": "amIACat",
          "parameters": {
            "sound": "meow,nyaa"
          },
          "constraints": [
            {
              "contextName": "environment",
              "operator": "IN",
              "values": [
                "staging",
                "prod"
              ]
            }
          ]
        }
      ],
      "createdAt": "2018-10-13T10:15:29.009Z"
    },
    {
      "name": "CustomToggleWarning",
      "description": "CustomToggle Warning Test",
      "enabled": True,
      "strategies": [
        {
          "name": "amIADog",
          "parameters": {
            "sound": "arf,bark"
          }
        }
      ],
      "createdAt": "2018-10-13T10:15:29.009Z"
    },
    {
      "name": "CustomToggleWarningMultiStrat",
      "description": "CustomToggle Warning Test",
      "enabled": True,
      "strategies": [
        {
          "name": "amIADog",
          "parameters": {
            "sound": "arf,bark"
          }
        },
        {
          "name": "default",
          "parameters": {}
        }
      ],
      "createdAt": "2018-10-13T10:15:29.009Z"
    },
    {
      "name": "UserWithId",
      "description": "UserWithId",
      "enabled": True,
      "strategies": [
        {
          "name": "userWithId",
          "parameters": {
            "userIds": "meep@meep.com,test@test.com,ivan@ivan.com"
          }
        }
      ],
      "createdAt": "2018-10-11T09:33:51.171Z"
    }
  ]
}



# pylint: disable=invalid-name,dangerous-default-value
from typing import Iterator
import warnings
from UnleashClient.constraints import Constraint


class Strategy:
    """
    The parent class for default and custom strategies.

    In general, default & custom classes should only need to override:

    - ``__init__()`` - Depending on the parameters your feature needs
    - ``apply()`` - Your feature provisioning

    :param constraints: List of 'constraints' objects derived from strategy section (...from feature section) of `/api/clients/features` Unleash server response.
    :param parameters: The 'parameter' objects from the strategy section (...from feature section) of `/api/clients/features` Unleash server response.
    """

    def __init__(self,
                 constraints: list = [],
                 parameters: dict = {},
                 segment_ids: list = None,
                 global_segments: dict = None
                 ) -> None:
        self.parameters = parameters
        self.constraints = constraints
        self.segment_ids = segment_ids or []
        self.global_segments = global_segments or {}
        self.parsed_provisioning = self.load_provisioning()

    def __call__(self, context: dict = None):
        warnings.warn(
            "unleash-client-python v3.x.x requires overriding the execute() method instead of the __call__() method.",
            DeprecationWarning
        )

    def execute(self, context: dict = None) -> bool:
        """
        Executes the strategies by:

        - Checking constraints
        - Applying the strategy

        This is what UnleashClient calls when you run ``is_enabled()``

        :param context: Feature flag context.
        :return: Feature flag result.
        """
        flag_state = False

        if all(constraint.apply(context) for constraint in self.parsed_constraints):
            flag_state = self.apply(context)

        return flag_state

    @property
    def parsed_constraints(self) -> Iterator[Constraint]:
        for constraint_dict in self.constraints:
            yield Constraint(constraint_dict=constraint_dict)

        for segment_id in self.segment_ids:
            segment = self.global_segments[segment_id]
            for constraint in segment["constraints"]:
                yield Constraint(constraint_dict=constraint)

    def load_provisioning(self) -> list:  # pylint: disable=no-self-use
        """
        Loads strategy provisioning from Unleash feature flag configuration.

        This should parse the raw values in ``self.parameters`` into format Python can comprehend.
        """
        return []

    def apply(self, context: dict = None) -> bool:  # pylint: disable=unused-argument,no-self-use
        """
        Strategy implementation.

        :param context: Feature flag context
        :return: Feature flag result
        """
        return False



# pylint: disable=invalid-name
import ipaddress
from UnleashClient.strategies.Strategy import Strategy
from UnleashClient.utils import LOGGER


class RemoteAddress(Strategy):
    def load_provisioning(self) -> list:
        parsed_ips = []

        for address in self.parameters["IPs"].split(','):

            if "/" in address:
                try:
                    parsed_ips.append(ipaddress.ip_network(address.strip(), strict=True))  # type: ignore
                except (ipaddress.AddressValueError, ipaddress.NetmaskValueError, ValueError) as parsing_error:
                    LOGGER.warning("Error parsing IP range: %s", parsing_error)
            else:
                try:
                    parsed_ips.append(ipaddress.ip_address(address.strip()))  # type: ignore
                except (ipaddress.AddressValueError, ipaddress.NetmaskValueError, ValueError) as parsing_error:
                    LOGGER.warning("Error parsing IP : %s", parsing_error)

        return parsed_ips

    def apply(self, context: dict = None) -> bool:
        """
        Returns true if IP is in list of IPs

        :return:
        """
        return_value = False

        try:
            context_ip = ipaddress.ip_address(context["remoteAddress"])
        except (ipaddress.AddressValueError, ipaddress.NetmaskValueError, ValueError) as parsing_error:
            LOGGER.warning("Error parsing IP : %s", parsing_error)
            context_ip = None

        if context_ip:
            for addr_or_range in [value for value in self.parsed_provisioning if value.version == context_ip.version]:
                if isinstance(addr_or_range, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
                    if context_ip == addr_or_range:
                        return_value = True
                        break
                else:
                    if context_ip in addr_or_range:
                        return_value = True
                        break

        return return_value



import uuid
import pytest
from UnleashClient.strategies import GradualRolloutSessionId


def generate_context():
    return {"sessionId": uuid.uuid4()}


@pytest.fixture()
def strategy():
    yield GradualRolloutSessionId(parameters={"percentage": 50, "groupId": "test"})


def test_userwithid(strategy):
    strategy.execute(context=generate_context())



import time
from UnleashClient import UnleashClient


# ---
import logging
import sys

root = logging.getLogger()
root.setLevel(logging.DEBUG)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
root.addHandler(handler)
# ---

my_client = UnleashClient(
    url="https://unleash.herokuapp.com/api",
    environment="staging",
    app_name="pyIvan",
)

my_client.initialize_client()

while True:
    time.sleep(10)
    context = {
        'userId': "1",
        'sound': 'woof'
    }
    print(f"ivantest: {my_client.is_enabled('ivantest', context)}")



import pytest
from UnleashClient.strategies import RemoteAddress
from tests.utilities.testing_constants import IP_LIST


@pytest.fixture()
def strategy():
    yield RemoteAddress(parameters={"IPs": IP_LIST})


def test_init_with_bad_address():
    BAD_IP_LIST = IP_LIST + ",garbage"
    strategy = RemoteAddress(parameters={"IPs": BAD_IP_LIST})
    assert len(strategy.parsed_provisioning) == 4


def test_init_with_bad_range():
    BAD_IP_LIST = IP_LIST + ",ga/rbage"
    strategy = RemoteAddress(parameters={"IPs": BAD_IP_LIST})
    assert len(strategy.parsed_provisioning) == 4


def test_ipv4_range(strategy):
    assert strategy.execute(context={"remoteAddress": "69.208.0.1"})


def test_ipv4_value(strategy):
    assert strategy.execute(context={"remoteAddress": "70.208.1.1"})


def test_ipv6_rangee(strategy):
    assert strategy.execute(context={"remoteAddress": "2001:db8:1234:0000:0000:0000:0000:0001"})


def test_ipv6_value(strategy):
    assert strategy.execute(context={"remoteAddress": "2002:db8:1234:0000:0000:0000:0000:0001"})


def test_garbage_value(strategy):
    assert not strategy.execute(context={"remoteAddress": "WTFISTHISURCRAZY"})



import copy
from UnleashClient.loader import load_features
from UnleashClient.features import Feature
from UnleashClient.strategies import GradualRolloutUserId, FlexibleRollout, UserWithId
from UnleashClient.variants import Variants
from UnleashClient.constants import FEATURES_URL, FAILED_STRATEGIES
from tests.utilities.mocks import MOCK_ALL_FEATURES
from tests.utilities.testing_constants import DEFAULT_STRATEGY_MAPPING


def test_loader_initialization(cache_full):  # noqa: F811
    # Set up variables
    in_memory_features = {}
    temp_cache = cache_full

    # Tests
    load_features(temp_cache, in_memory_features, DEFAULT_STRATEGY_MAPPING)
    assert isinstance(in_memory_features["GradualRolloutUserID"], Feature)
    assert isinstance(in_memory_features["GradualRolloutUserID"].strategies[0], GradualRolloutUserId)

    for feature_name in in_memory_features.keys():
        if feature_name == 'Garbage':  # Don't check purposely invalid strategy.
            break

        feature = in_memory_features[feature_name]
        assert len(feature.strategies) > 0
        strategy = feature.strategies[0]

        if isinstance(strategy, UserWithId):
            assert strategy.parameters
            assert len(strategy.parsed_provisioning)

        if isinstance(strategy, FlexibleRollout):
            len(list(strategy.parsed_constraints)) > 0

        if isinstance(strategy, Variants):
            assert strategy.variants


def test_loader_refresh_strategies(cache_full):  # noqa: F811
    # Set up variables
    in_memory_features = {}
    temp_cache = cache_full

    load_features(temp_cache, in_memory_features, DEFAULT_STRATEGY_MAPPING)

    # Simulate update mutation
    mock_updated = copy.deepcopy(MOCK_ALL_FEATURES)
    mock_updated["features"][4]["strategies"][0]["parameters"]["percentage"] = 60
    temp_cache.set(FEATURES_URL, mock_updated)

    load_features(temp_cache, in_memory_features, DEFAULT_STRATEGY_MAPPING)

    assert in_memory_features["GradualRolloutUserID"].strategies[0].parameters["percentage"] == 60
    assert len(temp_cache.get(FAILED_STRATEGIES)) == 1


def test_loader_refresh_variants(cache_full):  # noqa: F811
    # Set up variables
    in_memory_features = {}
    temp_cache = cache_full

    load_features(temp_cache, in_memory_features, DEFAULT_STRATEGY_MAPPING)

    # Simulate update mutation
    mock_updated = copy.deepcopy(MOCK_ALL_FEATURES)
    mock_updated["features"][8]["variants"][0]["name"] = "VariantA"
    temp_cache.set(FEATURES_URL, mock_updated)

    load_features(temp_cache, in_memory_features, DEFAULT_STRATEGY_MAPPING)

    assert in_memory_features["Variations"].variants.variants[0]["name"] == "VariantA"


def test_loader_initialization_failure(cache_custom):  # noqa: F811
    # Set up variables
    in_memory_features = {}
    temp_cache = cache_custom

    # Tests
    load_features(temp_cache, in_memory_features, DEFAULT_STRATEGY_MAPPING)
    assert isinstance(in_memory_features["UserWithId"], Feature)


def test_loader_segments(cache_segments):
    # Set up variables
    in_memory_features = {}
    temp_cache = cache_segments

    load_features(temp_cache, in_memory_features, DEFAULT_STRATEGY_MAPPING)
    feature = in_memory_features["Test"]
    loaded_constraints = list(feature.strategies[0].parsed_constraints)
    assert(len(loaded_constraints) == 2)



import uuid
import json
import pytest
import responses
from UnleashClient import UnleashClient
from UnleashClient.constants import FEATURES_URL, METRICS_URL, REGISTER_URL
from tests.utilities.testing_constants import URL, APP_NAME


MOCK_JSON = """
{
    "version": 1,
    "features": [
        {
            "name": "Feature.constraints.no_values",
            "description": "Not enabled with constraints and no values",
            "enabled": true,
            "strategies": [
                {
                    "name": "default",
                    "parameters": {},
                    "constraints": [
                        {
                            "contextName": "environment",
                            "operator": "IN",
                            "values": []

                        }
                    ]
                }
            ]
        },
        {
            "name": "Feature.constraints.no_values_NOT_IN",
            "description": "Is enabled with constraints and NOT_IN empty values",
            "enabled": true,
            "strategies": [
                {
                    "name": "default",
                    "parameters": {},
                    "constraints": [
                        {
                            "contextName": "environment",
                            "operator": "NOT_IN",
                            "values": []

                        }
                    ]
                }
            ]
        },
        {
            "name": "Feature.constraints.empty",
            "description": "Is enabled with empty constraints array",
            "enabled": true,
            "strategies": [
                {
                    "name": "default",
                    "parameters": {},
                    "constraints": []
                }
            ]
        }
    ]
}
"""


@pytest.fixture()
def unleash_client():
    unleash_client = UnleashClient(url=URL,
                                   app_name=APP_NAME,
                                   instance_id='pytest_%s' % uuid.uuid4())
    yield unleash_client
    unleash_client.destroy()


@responses.activate
def test_feature_constraints_novalue_defaultenv(unleash_client):
    """
    Feature.constraints.no_value should not be enabled in default environment
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.constraints.no_values", {'environment': 'default'})


@responses.activate
def test_feature_constraints_novalue_emptyenv(unleash_client):
    """
    Feature.constraints.no_value should not be enabled in empty environment
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.constraints.no_values", {})


@responses.activate
def test_feature_constraints_novalue_defaultenv_notin(unleash_client):
    """
    Feature.constraints.no_values_NOT_IN should be enabled in default environment
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.constraints.no_values_NOT_IN", {'environment': 'default'})


@responses.activate
def test_feature_constraints_novalue_emptyenv_notin(unleash_client):
    """
    Feature.constraints.no_values_NOT_IN should be enabled in empty environment
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.constraints.no_values_NOT_IN", {})


@responses.activate
def test_feature_constraints_empty_defaultenv(unleash_client):
    """
    Feature.constraints.empty should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.constraints.no_values_NOT_IN", {'environment': 'default'})


@responses.activate
def test_feature_constraints_empty_emptyenv(unleash_client):
    """
    Feature.constraints.empty should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.constraints.no_values_NOT_IN", {})


import time
from UnleashClient import UnleashClient

# ---
import logging
import sys

root = logging.getLogger()
root.setLevel(logging.DEBUG)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
root.addHandler(handler)
# ---

my_client = UnleashClient(
    url="http://localhost:4242/api",
    app_name="pyIvan"
)

my_client.initialize_client()

while True:
    time.sleep(10)
    print(my_client.is_enabled("Demo"))



import pytest
from UnleashClient.strategies import GradualRolloutRandom


@pytest.fixture()
def strategy():
    yield GradualRolloutRandom(parameters={"percentage": 50})


def test_userwithid(strategy):
    assert isinstance(strategy.execute(), bool)



import pytest
from tests.utilities import generate_email_list
from UnleashClient.strategies import UserWithId

(EMAIL_LIST, CONTEXT) = generate_email_list(20)


@pytest.fixture()
def strategy():
    yield UserWithId(parameters={"userIds": EMAIL_LIST})


def test_userwithid(strategy):
    assert strategy.execute(context=CONTEXT)


def test_userwithid_missing_parameter(strategy):
    assert not strategy.execute(context={})



CONSTRAINT_DICT_IN = \
    {
        "contextName": "appName",
        "operator": "IN",
        "values": [
            "test",
            "test2"
        ]
    }


CONSTRAINT_DICT_NOTIN = \
    {
        "contextName": "appName",
        "operator": "NOT_IN",
        "values": [
            "test",
            "test2"
        ]
    }


CONSTRAINT_DICT_STR_INVERT = \
    {
        "contextName": "customField",
        "operator": "STR_CONTAINS",
        "values": ["dog", "cat", "hAmStEr"],
        "caseInsensitive": True,
        "inverted": True
    }

CONSTRAINT_DICT_STR_CONTAINS_CI = \
    {
        "contextName": "customField",
        "operator": "STR_CONTAINS",
        "values": ["dog", "cat", "hAmStEr"],
        "caseInsensitive": True,
        "inverted": False
    }


CONSTRAINT_DICT_STR_CONTAINS_NOT_CI = \
    {
        "contextName": "customField",
        "operator": "STR_CONTAINS",
        "values": ["dog", "cat", "hAmStEr"],
        "caseInsensitive": False,
        "inverted": False
    }


CONSTRAINT_DICT_STR_ENDS_WITH_CI = \
    {
        "contextName": "customField",
        "operator": "STR_ENDS_WITH",
        "values": ["dog", "cat", "hAmStEr"],
        "caseInsensitive": True,
        "inverted": False
    }

CONSTRAINT_DICT_STR_ENDS_WITH_NOT_CI = \
    {
        "contextName": "customField",
        "operator": "STR_ENDS_WITH",
        "values": ["dog", "cat", "hAmStEr"],
        "caseInsensitive": False,
        "inverted": False
    }


CONSTRAINT_DICT_STR_STARTS_WITH_CI = \
    {
        "contextName": "customField",
        "operator": "STR_STARTS_WITH",
        "values": ["dog", "cat", "hAmStEr"],
        "caseInsensitive": True,
        "inverted": False
    }


CONSTRAINT_DICT_STR_STARTS_WITH_NOT_CI = \
    {
        "contextName": "customField",
        "operator": "STR_STARTS_WITH",
        "values": ["dog", "cat", "hAmStEr"],
        "caseInsensitive": False,
        "inverted": False
    }


CONSTRAINT_NUM_EQ = \
    {
        "contextName": "customField",
        "operator": "NUM_EQ",
        "value": "5",
        "inverted": False
    }


CONSTRAINT_NUM_GT = \
    {
        "contextName": "customField",
        "operator": "NUM_GT",
        "value": "5",
        "inverted": False
    }


CONSTRAINT_NUM_GTE = \
    {
        "contextName": "customField",
        "operator": "NUM_GTE",
        "value": 5,
        "inverted": False
    }


CONSTRAINT_NUM_LT = \
    {
        "contextName": "customField",
        "operator": "NUM_LT",
        "value": "5",
        "inverted": False
    }


CONSTRAINT_NUM_LTE = \
    {
        "contextName": "customField",
        "operator": "NUM_LTE",
        "value": "5",
        "inverted": False
    }


CONSTRAINT_NUM_FLOAT = \
    {
        "contextName": "customField",
        "operator": "NUM_LTE",
        "value": "5.1",
        "inverted": False
    }


CONSTRAINT_DATE_AFTER = \
    {
        "contextName": "currentTime",
        "operator": "DATE_AFTER",
        "value": "2022-01-22T00:00:00.000Z",
        "inverted": False
    }


CONSTRAINT_DATE_BEFORE = \
    {
        "contextName": "currentTime",
        "operator": "DATE_BEFORE",
        "value": "2022-01-22T00:00:00.000Z",
        "inverted": False
    }


CONSTRAINT_DATE_ERROR = \
    {
        "contextName": "currentTime",
        "operator": "DATE_AFTER",
        "value": "abcd",
        "inverted": False
    }


CONSTRAINT_SEMVER_EQ = \
    {
        "contextName": "customField",
        "operator": "SEMVER_EQ",
        "value": "1.2.2",
        "inverted": False
    }


CONSTRAINT_SEMVER_GT = \
    {
        "contextName": "customField",
        "operator": "SEMVER_GT",
        "value": "1.2.2",
        "inverted": False
    }


CONSTRAINT_SEMVER_LT = \
    {
        "contextName": "customField",
        "operator": "SEMVER_LT",
        "value": "1.2.2",
        "inverted": False
    }



import pytest
from UnleashClient.strategies.FlexibleRolloutStrategy import FlexibleRollout

BASE_FLEXIBLE_ROLLOUT_DICT = \
    {
        "name": "flexibleRollout",
        "parameters": {
            "rollout": 50,
            "stickiness": "userId",
            "groupId": "AB12A"
        },
        "constraints": [
            {
                "contextName": "environment",
                "operator": "IN",
                "values": [
                    "staging",
                    "prod"
                ]
            },
            {
                "contextName": "userId",
                "operator": "IN",
                "values": [
                    "122",
                    "155",
                    "9"
                ]
            },
            {
                "contextName": "userId",
                "operator": "NOT_IN",
                "values": [
                    "4"
                ]
            },
            {
                "contextName": "appName",
                "operator": "IN",
                "values": [
                    "test"
                ]
            }
        ]
    }


@pytest.fixture()
def strategy():
    yield FlexibleRollout(BASE_FLEXIBLE_ROLLOUT_DICT['constraints'], BASE_FLEXIBLE_ROLLOUT_DICT['parameters'])


def test_flexiblerollout_satisfiesconstraints(strategy):
    context = {
        'userId': "122",
        'appName': 'test',
        'environment': 'prod'
    }

    assert strategy.execute(context)


def test_flexiblerollout_doesntsatisfiesconstraints(strategy):
    context = {
        'userId': "2",
        'appName': 'qualityhamster',
        'environment': 'prod'
    }
    assert not strategy.execute(context)


def test_flexiblerollout_userid(strategy):
    base_context = dict(appName='test', environment='prod')
    base_context['userId'] = "122"
    assert strategy.execute(base_context)
    base_context['userId'] = "155"
    assert not strategy.execute(base_context)


def test_flexiblerollout_sessionid(strategy):
    BASE_FLEXIBLE_ROLLOUT_DICT['parameters']['stickiness'] = 'sessionId'
    base_context = dict(appName='test', environment='prod', userId="9")
    base_context['sessionId'] = "122"
    assert strategy.execute(base_context)
    base_context['sessionId'] = "155"
    assert not strategy.execute(base_context)


def test_flexiblerollout_random(strategy):
    BASE_FLEXIBLE_ROLLOUT_DICT['parameters']['stickiness'] = 'random'
    base_context = dict(appName='test', environment='prod', userId="1")
    assert strategy.execute(base_context) in [True, False]


def test_flexiblerollout_customfield(strategy):
    BASE_FLEXIBLE_ROLLOUT_DICT['parameters']['stickiness'] = 'customField'
    base_context = dict(appName='test', environment='prod', userId="9")
    base_context['customField'] = "122"
    assert strategy.execute(base_context)
    base_context['customField'] = "155"
    assert not strategy.execute(base_context)


def test_flexiblerollout_default():
    BASE_FLEXIBLE_ROLLOUT_DICT['parameters']['stickiness'] = 'default'
    BASE_FLEXIBLE_ROLLOUT_DICT['constraints'] = [x for x in BASE_FLEXIBLE_ROLLOUT_DICT['constraints'] if x['contextName'] != 'userId']
    strategy = FlexibleRollout(BASE_FLEXIBLE_ROLLOUT_DICT['constraints'], BASE_FLEXIBLE_ROLLOUT_DICT['parameters'])
    base_context = dict(appName='test', environment='prod', userId="122", sessionId="155")
    assert strategy.execute(base_context)
    base_context = dict(appName='test', environment='prod', sessionId="122")
    assert strategy.execute(base_context)
    base_context = dict(appName='test', environment='prod')
    assert strategy.execute(base_context) in [True, False]



import uuid
import json
import pytest
import responses
from UnleashClient import UnleashClient
from UnleashClient.constants import FEATURES_URL, METRICS_URL, REGISTER_URL
from tests.utilities.testing_constants import URL, APP_NAME


MOCK_JSON = """
 {
    "version": 1,
    "features": [
        {
            "name": "Feature.flexible.rollout.custom.stickiness_100",
            "description": "Should support custom stickiness as option",
            "enabled": true,
            "strategies": [
                {
                    "name": "flexibleRollout",
                    "parameters": {
                        "rollout": "100",
                        "stickiness": "customField",
                        "groupId": "Feature.flexible.rollout.custom.stickiness_100"
                    },
                    "constraints": []
                }
            ],
            "variants": [
                {
                    "name": "blue",
                    "weight": 25,
                    "stickiness": "customField",
                    "payload": {
                        "type": "string",
                        "value": "val1"
                    }
                },
                {
                    "name": "red",
                    "weight": 25,
                    "stickiness": "customField",
                    "payload": {
                        "type": "string",
                        "value": "val1"
                    }
                },
                {
                    "name": "green",
                    "weight": 25,
                    "stickiness": "customField",
                    "payload": {
                        "type": "string",
                        "value": "val1"
                    }
                },
                {
                    "name": "yellow",
                    "weight": 25,
                    "stickiness": "customField",
                    "payload": {
                        "type": "string",
                        "value": "val1"
                    }
                }
            ]
        },
        {
            "name": "Feature.flexible.rollout.custom.stickiness_50",
            "description": "Should support custom stickiness as option",
            "enabled": true,
            "strategies": [
                {
                    "name": "flexibleRollout",
                    "parameters": {
                        "rollout": "50",
                        "stickiness": "customField",
                        "groupId": "Feature.flexible.rollout.custom.stickiness_50"
                    },
                    "constraints": []
                }
            ]
        }
    ]
}
"""


@pytest.fixture()
def unleash_client():
    unleash_client = UnleashClient(url=URL,
                                   app_name=APP_NAME,
                                   instance_id='pytest_%s' % uuid.uuid4())
    yield unleash_client
    unleash_client.destroy()


@responses.activate
def test_feature_flexiblerollout_stickiness_100(unleash_client):
    """
    Feature.flexible.rollout.custom.stickiness_100 should be enabled without field defined for 100%
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.flexible.rollout.custom.stickiness_100", {'customField': 'any_value'})


@responses.activate
def test_feature_flexiblerollout_stickiness_50_nocontext(unleash_client):
    """
    Feature.flexible.rollout.custom.stickiness_50 should not be enabled without custom field
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.flexible.rollout.custom.stickiness_50", {})


@responses.activate
def test_feature_flexiblerollout_stickiness_50_customfield_402(unleash_client):
    """
    Feature.flexible.rollout.custom.stickiness_50 should not enabled without customField=402
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.flexible.rollout.custom.stickiness_50", {'customField': '402'})


@responses.activate
def test_feature_flexiblerollout_stickiness_50_customfield_388(unleash_client):
    """
    Feature.flexible.rollout.custom.stickiness_50 should be enabled for customField=388
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.flexible.rollout.custom.stickiness_50", {'customField': '388'})


@responses.activate
def test_feature_flexiblerollout_stickiness_50_customfield_39(unleash_client):
    """
    Feature.flexible.rollout.custom.stickiness_50 should be enabled without customField=39
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.flexible.rollout.custom.stickiness_50", {'customField': '39'})


@responses.activate
def test_variant_flexiblerollout_stickiness_100_customfield_528(unleash_client):
    """
    Feature.flexible.rollout.custom.stickiness_100 and customField=528 yields blue
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'customField': "528"
    }

    expected_result = {
        "name": "blue",
        "payload": {
            "type": "string",
            "value": "val1"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.flexible.rollout.custom.stickiness_100", context)
    assert actual_result == expected_result


@responses.activate
def test_variant_flexiblerollout_stickiness_100_customfield_16(unleash_client):
    """
    Feature.flexible.rollout.custom.stickiness_100 and customField=16 yields blue
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'customField': "16"
    }

    expected_result = {
        "name": "blue",
        "payload": {
            "type": "string",
            "value": "val1"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.flexible.rollout.custom.stickiness_100", context)
    assert actual_result == expected_result


@responses.activate
def test_variant_flexiblerollout_stickiness_100_customfield_198(unleash_client):
    """
    Feature.flexible.rollout.custom.stickiness_100 and customField=198 yields red
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'customField': "198"
    }

    expected_result = {
        "name": "red",
        "payload": {
            "type": "string",
            "value": "val1"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.flexible.rollout.custom.stickiness_100", context)
    assert actual_result == expected_result


@responses.activate
def test_variant_flexiblerollout_stickiness_100_customfield_43(unleash_client):
    """
    Feature.flexible.rollout.custom.stickiness_100 and customField=43 yields green
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'customField': "43"
    }

    expected_result = {
        "name": "green",
        "payload": {
            "type": "string",
            "value": "val1"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.flexible.rollout.custom.stickiness_100", context)
    assert actual_result == expected_result


@responses.activate
def test_variant_flexiblerollout_stickiness_100_customfield_112(unleash_client):
    """
    Feature.flexible.rollout.custom.stickiness_100 and customField=112 yields yellow
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'customField': "112"
    }

    expected_result = {
        "name": "yellow",
        "payload": {
            "type": "string",
            "value": "val1"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.flexible.rollout.custom.stickiness_100", context)
    assert actual_result == expected_result



import json
import requests
from UnleashClient.constants import REQUEST_TIMEOUT, APPLICATION_HEADERS, METRICS_URL
from UnleashClient.utils import LOGGER, log_resp_info


# pylint: disable=broad-except
def send_metrics(url: str,
                 request_body: dict,
                 custom_headers: dict,
                 custom_options: dict) -> bool:
    """
    Attempts to send metrics to Unleash server

    Notes:
    * If unsuccessful (i.e. not HTTP status code 200), message will be logged

    :param url:
    :param request_body:
    :param custom_headers:
    :param custom_options:
    :return: true if registration successful, false if registration unsuccessful or exception.
    """
    try:
        LOGGER.info("Sending messages to with unleash @ %s", url)
        LOGGER.info("unleash metrics information: %s", request_body)

        resp = requests.post(url + METRICS_URL,
                             data=json.dumps(request_body),
                             headers={**custom_headers, **APPLICATION_HEADERS},
                             timeout=REQUEST_TIMEOUT, **custom_options)

        if resp.status_code != 202:
            log_resp_info(resp)
            LOGGER.warning("Unleash CLient metrics submission failed.")
            return False

        LOGGER.info("Unleash Client metrics successfully sent!")

        return True
    except requests.RequestException as exc:
        LOGGER.warning("Unleash Client metrics submission failed due to exception: %s", exc)

    return False



# pylint: disable=invalid-name
from UnleashClient.utils import normalized_hash
from UnleashClient.strategies.Strategy import Strategy


class GradualRolloutUserId(Strategy):
    def apply(self, context: dict = None) -> bool:
        """
        Returns true if userId is a member of id list.

        :return:
        """
        percentage = int(self.parameters["percentage"])
        activation_group = self.parameters["groupId"]

        return percentage > 0 and normalized_hash(context["userId"], activation_group) <= percentage



import time
from UnleashClient import UnleashClient

# ---
import logging
import sys

root = logging.getLogger()
root.setLevel(logging.DEBUG)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
root.addHandler(handler)
# ---


my_client = UnleashClient(
    url="https://gitlab.com/api/v4/feature_flags/unleash/32635317",
    app_name="pyIvan",
    instance_id="Sc5fv9aCyFPB4XcEFk-E",
    disable_metrics=True,
    disable_registration=True
)

my_client.initialize_client()

while True:
    time.sleep(10)
    print(my_client.is_enabled("test"))



MOCK_ALL_FEATURES = \
{
  "version": 1,
  "features": [
    {
      "name": "ApplicationHostname",
      "description": "Application Hostname strategy",
      "enabled": True,
      "strategies": [
        {
          "name": "applicationHostname",
          "parameters": {
            "hostNames": "iMacPro.local,test1,test2"
          }
        }
      ],
      "createdAt": "2018-10-09T06:05:14.757Z"
    },
    {
      "name": "Default",
      "description": "Default feature toggle",
      "enabled": True,
      "strategies": [
        {
          "name": "default",
          "parameters": {}
        }
      ],
      "createdAt": "2018-10-09T06:04:05.667Z"
    },
    {
      "name": "GradualRolloutRandom",
      "description": "Gradual Rollout Random example",
      "enabled": True,
      "strategies": [
        {
          "name": "gradualRolloutRandom",
          "parameters": {
            "percentage": 50
          }
        }
      ],
      "createdAt": "2018-10-09T06:05:37.637Z"
    },
    {
      "name": "GradualRolloutSessionId",
      "description": "SessionID check!",
      "enabled": True,
      "strategies": [
        {
          "name": "gradualRolloutSessionId",
          "parameters": {
            "percentage": 50,
            "groupId": "GradualRolloutSessionId"
          }
        }
      ],
      "createdAt": "2018-10-09T06:06:51.057Z"
    },
    {
      "name": "GradualRolloutUserID",
      "description": "GradualRolloutUserID strategy",
      "enabled": True,
      "strategies": [
        {
          "name": "gradualRolloutUserId",
          "parameters": {
            "percentage": 50,
            "groupId": "GradualRolloutUserID"
          }
        }
      ],
      "createdAt": "2018-10-09T06:07:17.520Z"
    },
    {
      "name": "RemoteAddress",
      "description": "RemoteAddress strategies",
      "enabled": True,
      "strategies": [
        {
          "name": "remoteAddress",
          "parameters": {
            "IPs": "69.208.0.0/29,70.208.1.1,2001:db8:1234::/48,2002:db8:1234:0000:0000:0000:0000:0001"
          }
        }
      ],
      "createdAt": "2018-10-09T06:08:42.398Z"
    },
    {
      "name": "UserWithId",
      "description": "UserWithId strategies",
      "enabled": True,
      "strategies": [
        {
          "name": "userWithId",
          "parameters": {
            "userIds": "meep@meep.com,test@test.com,wat@wat.com"
          }
        }
      ],
      "createdAt": "2018-10-09T06:09:19.203Z"
    },
    {
      "name": "FlexibleRollout",
      "description": "FlexibleRollout strategies",
      "enabled": True,
      "strategies": [
        {
          "name": "flexibleRollout",
          "parameters": {
            "rollout": "21",
            "stickiness": "userId",
            "groupId": "ivantest"
          },
          "constraints": [
            {
              "contextName": "environment",
              "operator": "IN",
              "values": [
                "staging",
                "prod"
              ]
            },
            {
              "contextName": "userId",
              "operator": "NOT_IN",
              "values": [
                "1",
                "2",
                "3"
              ]
            },
            {
              "contextName": "userId",
              "operator": "IN",
              "values": [
                "4",
                "5",
                "6"
              ]
            },
            {
              "contextName": "appName",
              "operator": "IN",
              "values": [
                "test"
              ]
            }
          ]
        }
      ],
      "variants": None,
      "createdAt": "2019-10-05T07:30:29.896Z"
    },
    {
      "name": "Variations",
      "description": "Test variation",
      "enabled": True,
      "strategies": [
          {
              "name": "default"
          }
      ],
      "variants": [
          {
              "name": "VarA",
              "weight": 34,
              "payload": {
                  "type": "string",
                  "value": "Test1"
              },
              "overrides": [
                  {
                      "contextName": "userId",
                      "values": [
                          "ivanklee86@gmail.com",
                          "ivan@aaptiv.com"
                      ]
                  }
              ]
          },
          {
              "name": "VarB",
              "weight": 33,
              "payload": {
                  "type": "string",
                  "value": "Test 2"
              }
          },
          {
              "name": "VarC",
              "weight": 33,
              "payload": {
                  "type": "string",
                  "value": "Test 3"
              }
          }
      ],
      "createdAt": "2019-10-25T13:22:02.035Z"
  },
  {
    "name": "Garbage",
    "description": "Invalid strategy",
    "enabled": True,
    "strategies": [
      {
        "name": "blargwatisdis",
        "parameters": {}
      }
    ],
    "createdAt": "2018-10-09T06:04:05.667Z"
  },
  ]
}



"""Setup file for UnleashClient"""
from setuptools import setup, find_packages


def readme():
    """Include README.rst content in PyPi build information"""
    with open('README.md') as file:
        return file.read()


setup(
    name='UnleashClient',
    author='Ivan Lee',
    author_email='ivanklee86@gmail.com',
    description='Python client for the Unleash feature toggle system!',
    long_description=readme(),
    long_description_content_type="text/markdown",
    url='https://github.com/Unleash/unleash-client-python',
    packages=find_packages(exclude=["tests*"]),
    package_data={"UnleashClient": ["py.typed"]},
    install_requires=[
        "requests",
        "fcache",
        "mmh3",
        "apscheduler",
        "importlib_metadata",
        "python-dateutil",
        "semver < 3.0.0"
    ],
    setup_requires=['setuptools_scm'],
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ]
)



import pytest
from UnleashClient.strategies import Default


@pytest.fixture()
def strategy():
    yield Default()


def test_defaultstrategy(strategy):
    assert isinstance(strategy.execute(), bool)



# pylint: disable=invalid-name
from UnleashClient.strategies.Strategy import Strategy


class Default(Strategy):
    def apply(self, context: dict = None) -> bool:
        """
        Return true if enabled.

        :return:
        """
        return True



import responses
from pytest import mark, param

from tests.utilities.mocks.mock_features import MOCK_FEATURE_RESPONSE, MOCK_FEATURE_RESPONSE_PROJECT
from tests.utilities.testing_constants import URL, APP_NAME, INSTANCE_ID, CUSTOM_HEADERS, CUSTOM_OPTIONS, PROJECT_URL, PROJECT_NAME, ETAG_VALUE
from UnleashClient.constants import FEATURES_URL
from UnleashClient.api import get_feature_toggles


FULL_FEATURE_URL = URL + FEATURES_URL


@responses.activate
@mark.parametrize("response,status,calls,expected", (
    param(MOCK_FEATURE_RESPONSE, 200, 1, lambda result: result["version"] == 1, id="success"),
    param(MOCK_FEATURE_RESPONSE, 202, 1, lambda result: not result, id="failure"),
    param({}, 500, 4, lambda result: not result, id="failure"),
))
def test_get_feature_toggle(response, status, calls, expected):
    responses.add(responses.GET, FULL_FEATURE_URL, json=response, status=status, headers={'etag': ETAG_VALUE})

    (result, etag) = get_feature_toggles(URL,
                                         APP_NAME,
                                         INSTANCE_ID,
                                         CUSTOM_HEADERS,
                                         CUSTOM_OPTIONS)

    assert len(responses.calls) == calls
    assert expected(result)


@responses.activate
def test_get_feature_toggle_project():
    responses.add(responses.GET, PROJECT_URL, json=MOCK_FEATURE_RESPONSE_PROJECT, status=200, headers={'etag': ETAG_VALUE})

    (result, etag) = get_feature_toggles(URL,
                                         APP_NAME,
                                         INSTANCE_ID,
                                         CUSTOM_HEADERS,
                                         CUSTOM_OPTIONS,
                                         PROJECT_NAME)

    assert len(responses.calls) == 1
    assert len(result["features"]) == 1
    assert etag == ETAG_VALUE


@responses.activate
def test_get_feature_toggle_failed_etag():
    responses.add(responses.GET, PROJECT_URL, json={}, status=500, headers={'etag': ETAG_VALUE})

    (result, etag) = get_feature_toggles(URL,
                                         APP_NAME,
                                         INSTANCE_ID,
                                         CUSTOM_HEADERS,
                                         CUSTOM_OPTIONS,
                                         PROJECT_NAME)

    assert len(responses.calls) == 4
    assert etag == ''


@responses.activate
def test_get_feature_toggle_etag_present():
    responses.add(responses.GET, PROJECT_URL, json={}, status=304, headers={'etag': ETAG_VALUE})

    (result, etag) = get_feature_toggles(URL,
                                         APP_NAME,
                                         INSTANCE_ID,
                                         CUSTOM_HEADERS,
                                         CUSTOM_OPTIONS,
                                         PROJECT_NAME,
                                         ETAG_VALUE)

    assert len(responses.calls) == 1
    assert not result
    assert responses.calls[0].request.headers['If-None-Match'] == ETAG_VALUE
    assert etag == ETAG_VALUE


@responses.activate
def test_get_feature_toggle_retries():
    responses.add(responses.GET, PROJECT_URL, json={}, status=500)
    responses.add(responses.GET, PROJECT_URL, json=MOCK_FEATURE_RESPONSE_PROJECT, status=200, headers={'etag': ETAG_VALUE})

    (result, etag) = get_feature_toggles(URL,
                                     APP_NAME,
                                     INSTANCE_ID,
                                     CUSTOM_HEADERS,
                                     CUSTOM_OPTIONS,
                                     PROJECT_NAME,
                                     ETAG_VALUE)

    assert len(responses.calls) == 2
    assert len(result["features"]) == 1
    assert etag == ETAG_VALUE
